# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from pygame_sdl2.error import error
import pygame_sdl2

@pygame_sdl2.register_init
def init():
    pygame_sdl2.display.sdl_main_init()

    if SDL_InitSubSystem(SDL_INIT_JOYSTICK):
        raise error()

@pygame_sdl2.register_quit
def quit(): # @ReservedAssignment
    SDL_QuitSubSystem(SDL_INIT_JOYSTICK)

def get_init():
    return SDL_WasInit(SDL_INIT_JOYSTICK) != 0

def get_count():
    return SDL_NumJoysticks()


cdef class Joystick:
    # Allow weak references.
    cdef object __weakref__

    cdef SDL_Joystick *joystick
    cdef int joyid

    def __cinit__(self):
        self.joystick = NULL

    def __init__(self, id):
        self.joyid = id

    def init(self):
        if self.joystick == NULL:
            self.joystick = SDL_JoystickOpen(self.joyid)
            if self.joystick == NULL:
                raise error()

    def quit(self): # @ReservedAssignment
        if self.joystick != NULL:
            SDL_JoystickClose(self.joystick)
            self.joystick = NULL

    def get_init(self):
        if self.joystick:
            return True
        else:
            return False

    def get_id(self):
        if self.joystick == NULL:
            raise error("joystick not initialized")

        return SDL_JoystickInstanceID(self.joystick)

    def get_name(self):
        cdef char *rv

        if self.joystick == NULL:
            raise error("joystick not initialized")

        rv = SDL_JoystickName(self.joystick)

        if rv:
            return rv
        else:
            return None

    def get_numaxes(self):
        if self.joystick == NULL:
            raise error("joystick not initialized")

        return SDL_JoystickNumAxes(self.joystick)

    def get_numballs(self):
        if self.joystick == NULL:
            raise error("joystick not initialized")

        return SDL_JoystickNumBalls(self.joystick)

    def get_numbuttons(self):
        if self.joystick == NULL:
            raise error("joystick not initialized")

        return SDL_JoystickNumButtons(self.joystick)

    def get_numhats(self):
        if self.joystick == NULL:
            raise error("joystick not initialized")

        return SDL_JoystickNumHats(self.joystick)

    def get_axis(self, axis_number):
        if self.joystick == NULL:
            raise error("joystick not initialized")

        return SDL_JoystickGetAxis(self.joystick, axis_number) / 32768.0

    def get_ball(self, ball_number):
        cdef int dx, dy

        if self.joystick == NULL:
            raise error("joystick not initialized")

        if SDL_JoystickGetBall(self.joystick, ball_number, &dx, &dy) == 0:
            return (dx, dy)
        else:
            raise error()

    def get_button(self, button):
        if self.joystick == NULL:
            raise error("joystick not initialized")

        return SDL_JoystickGetButton(self.joystick, button) == 1

    def get_hat(self, hat_number):
        if self.joystick == NULL:
            raise error("joystick not initialized")

        state = SDL_JoystickGetHat(self.joystick, hat_number)
        hx = hy = 0
        if state & SDL_HAT_LEFT:
            hx = -1
        elif state & SDL_HAT_RIGHT:
            hx = 1

        if state & SDL_HAT_UP:
            hy = 1
        elif state & SDL_HAT_DOWN:
            hy = -1

        return (hx, hy)



# Copyright 2014 Patrick Dawson <pat@dw.is>
# Copyright 2014 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from libc.string cimport memmove
from sdl2 cimport *

from pygame_sdl2.color cimport map_color, get_color
from pygame_sdl2.rect cimport to_sdl_rect
from pygame_sdl2.rect import Rect

from pygame_sdl2.error import error
from pygame_sdl2.locals import SRCALPHA
import pygame_sdl2

import warnings

cdef extern from "src/surface.h" nogil:
    int pygame_Blit (SDL_Surface * src, SDL_Rect * srcrect,
                 SDL_Surface * dst, SDL_Rect * dstrect, int the_args);

cdef void move_pixels(Uint8 *src, Uint8 *dst, int h, int span, int srcpitch, int dstpitch) nogil:
    if src < dst:
        src += (h - 1) * srcpitch;
        dst += (h - 1) * dstpitch;
        srcpitch = -srcpitch;
        dstpitch = -dstpitch;

    while h:
        h -= 1
        memmove(dst, src, span);
        src += srcpitch;
        dst += dstpitch;

# The total size of all allocated surfaces
total_size = 0

cdef class Surface:

    def __cinit__(self):
        self.surface = NULL
        self.owns_surface = False
        self.window_surface = False
        self.has_alpha = False

    def __dealloc__(self):
        global total_size

        if self.surface and self.owns_surface:
            if total_size:
                total_size -= self.surface.pitch * self.surface.h

            SDL_FreeSurface(self.surface)
            return
        elif self.window_surface:
            return
        elif self.parent:
            SDL_FreeSurface(self.surface)
            return

        warnings.warn("Memory leak via Surface in pygame_sdl2.")

    def __sizeof__(self):
        if self.surface and self.owns_surface:
            return self.surface.pitch * self.surface.h
        else:
            return 0

    def __init__(self, size, flags=0, depth=32, masks=None):

        self.locklist = None
        self.parent = None

        self.offset_x = 0
        self.offset_y = 0

        self.get_window_flags = None

        # When size is an empty tuple, we do not create a surface, and
        # expect our caller to set this object up.
        if size == ():
            return

        cdef int w
        cdef int h

        w, h = size
        assert w >= 0
        assert h >= 0

        cdef Uint32 Rmask, Gmask, Bmask, Amask
        cdef SDL_Surface *sample
        cdef Surface pysample
        cdef int depth_int

        if masks is not None:
            Rmask, Gmask, Bmask, Amask = masks
            depth_int = depth

        elif isinstance(depth, Surface):

            pysample = depth
            sample = pysample.surface
            Rmask = sample.format.Rmask
            Gmask = sample.format.Gmask
            Bmask = sample.format.Bmask
            Amask = sample.format.Amask
            depth_int = sample.format.BitsPerPixel

        else:

            pysample = pygame_sdl2.display.get_surface()

            if pysample and pysample.surface.format.BitsPerPixel == 32:
                sample = pysample.surface
                Rmask = sample.format.Rmask
                Gmask = sample.format.Gmask
                Bmask = sample.format.Bmask
                Amask = sample.format.Amask

            else:

                # RGB(A)
                if SDL_BYTEORDER == SDL_BIG_ENDIAN:
                    Rmask = 0xff000000
                    Gmask = 0x00ff0000
                    Bmask = 0x0000ff00
                    Amask = 0
                else:
                    Rmask = 0x000000ff
                    Gmask = 0x0000ff00
                    Bmask = 0x00ff0000
                    Amask = 0

            if (flags & SRCALPHA):
                if not Amask:
                    Amask = 0xffffffff & ~(Rmask | Gmask | Bmask)
            else:
                Amask = 0

            depth_int = 32

        cdef SDL_Surface *surface

        with nogil:
            surface = SDL_CreateRGBSurface(0, w, h, depth_int, Rmask, Gmask, Bmask, Amask)

        if not surface:
            raise error()

        self.take_surface(surface)

    cdef void take_surface(self, SDL_Surface *surface):
        if not surface:
            raise error("A null pointer was passed in.")

        self.surface = surface
        self.owns_surface = True

        global total_size

        total_size += self.surface.pitch * self.surface.h

    def __repr__(self):
        return "<Surface({}x{}x{})>".format(self.surface.w, self.surface.h, self.surface.format.BitsPerPixel)

    def blit(self, Surface source, dest, area=None, int special_flags=0):
        cdef SDL_Rect dest_rect
        cdef SDL_Rect area_rect
        cdef SDL_Rect *area_ptr = NULL

        cdef Surface temp

        cdef int err
        cdef Uint32 key
        cdef Uint8 alpha
        cdef bint colorkey

        colorkey = (SDL_GetColorKey(source.surface, &key) == 0)

        if not source.surface.format.Amask:
            if SDL_GetSurfaceAlphaMod(source.surface, &alpha):
                raise error()

            if alpha != 255 and (self.surface.format.Amask or colorkey):

                if area:
                    source = source.subsurface(area)
                    area = None

                if SDL_SetSurfaceBlendMode(source.surface, SDL_BLENDMODE_NONE):
                    raise error()
                temp = Surface(source.get_size(), SRCALPHA)

                with nogil:
                    err = SDL_UpperBlit(source.surface, NULL, temp.surface, NULL)

                if err:
                    raise error()

                source = temp
                colorkey = False

        if colorkey:
            if SDL_SetSurfaceBlendMode(source.surface, SDL_BLENDMODE_NONE):
                raise error()
        else:
            if SDL_SetSurfaceBlendMode(source.surface, SDL_BLENDMODE_BLEND):
                raise error()

        to_sdl_rect(dest, &dest_rect, "dest")

        if area is not None:
            to_sdl_rect(area, &area_rect, "area")
            area_ptr = &area_rect

        with nogil:

            if source.surface.format.Amask or special_flags:
                err = pygame_Blit(source.surface, area_ptr, self.surface, &dest_rect, special_flags)
            else:
                err = SDL_UpperBlit(source.surface, area_ptr, self.surface, &dest_rect)

        if err:
            raise error()

        dirty = Rect(dest[0], dest[1], source.surface.w, source.surface.h)
        return dirty.clip(self.get_rect())

    def convert(self, surface=None):
        if not isinstance(surface, Surface):
            surface = pygame_sdl2.display.get_surface()

        cdef SDL_PixelFormat *sample_format

        if surface is None:
            sample_format = SDL_AllocFormat(SDL_PIXELFORMAT_BGRX8888)
        else:
            sample_format = (<Surface> surface).surface.format

        cdef Uint32 amask
        cdef Uint32 rmask
        cdef Uint32 gmask
        cdef Uint32 bmask

        cdef Uint32 pixel_format
        cdef SDL_Surface *new_surface

        # If the sample surface has alpha, use it.
        if not sample_format.Amask:
            use_format = sample_format

            with nogil:
                new_surface = SDL_ConvertSurface(self.surface, sample_format, 0)

            if not new_surface:
                raise error()

        else:

            rmask = sample_format.Rmask
            gmask = sample_format.Gmask
            bmask = sample_format.Bmask
            amask = 0

            pixel_format = SDL_MasksToPixelFormatEnum(32, rmask, gmask, bmask, amask)

            with nogil:
                new_surface = SDL_ConvertSurfaceFormat(self.surface, pixel_format, 0)

            if not new_surface:
                raise error()

        cdef Surface rv = Surface(())
        rv.take_surface(new_surface)

        return rv

    def convert_alpha(self, Surface surface=None):
        if surface is None:
            surface = pygame_sdl2.display.get_surface()

        cdef SDL_PixelFormat *sample_format

        if surface is None:
            sample_format = SDL_AllocFormat(SDL_PIXELFORMAT_BGRA8888)
        else:
            sample_format = (<Surface> surface).surface.format

        cdef Uint32 amask = 0xff000000
        cdef Uint32 rmask = 0x00ff0000
        cdef Uint32 gmask = 0x0000ff00
        cdef Uint32 bmask = 0x000000ff

        cdef Uint32 pixel_format
        cdef SDL_Surface *new_surface

        # If the sample surface has alpha, use it.
        if sample_format.Amask:
            use_format = sample_format

            with nogil:
                new_surface = SDL_ConvertSurface(self.surface, sample_format, 0)

            if not new_surface:
                raise error()

        else:

            if sample_format.BytesPerPixel == 4:
                rmask = sample_format.Rmask
                gmask = sample_format.Gmask
                bmask = sample_format.Bmask
                amask = 0xffffffff & ~(rmask | gmask | bmask)

            pixel_format = SDL_MasksToPixelFormatEnum(32, rmask, gmask, bmask, amask)

            with nogil:
                new_surface = SDL_ConvertSurfaceFormat(self.surface, pixel_format, 0)

            if not new_surface:
                raise error()

        cdef Surface rv = Surface(())
        rv.take_surface(new_surface)

        return rv

    def copy(self):
        if self.surface.format.Amask:
            return self.convert_alpha(self)
        else:
            return self.convert(self)

    def fill(self, color, rect=None, special_flags=0):

        cdef SDL_Rect sdl_rect
        cdef Uint32 pixel = map_color(self.surface, color)
        cdef int err

        if rect is not None:
            to_sdl_rect(rect, &sdl_rect)

            if sdl_rect.x < 0:
                sdl_rect.w = sdl_rect.w + sdl_rect.x
                sdl_rect.x = 0

            if sdl_rect.y < 0:
                sdl_rect.w = sdl_rect.h + sdl_rect.y
                sdl_rect.y = 0

            if sdl_rect.w <= 0 or sdl_rect.h <= 0:
                return Rect(0, 0, 0, 0)

            with nogil:
                err = SDL_FillRect(self.surface, &sdl_rect, pixel)

            if err:
                raise error()

            return Rect(sdl_rect.x, sdl_rect.y, sdl_rect.w, sdl_rect.h)

        else:
            with nogil:
                err = SDL_FillRect(self.surface, NULL, pixel)

            if err:
                raise error()

            return Rect(0, 0, self.surface.w, self.surface.h)


    def scroll(self, int dx=0, int dy=0):
        cdef int srcx, destx, move_width
        cdef int srcy, desty, move_height

        cdef int width = self.surface.w
        cdef int height = self.surface.h

        cdef int per_pixel = self.surface.format.BytesPerPixel

        if dx >= 0:
            srcx = 0
            destx = dx
            move_width = width - dx
        else:
            srcx = -dx
            destx = 0
            move_width = width + dx

        if dy >= 0:
            srcy = 0
            desty = dy
            move_height = height - dy
        else:
            srcy = -dy
            desty = 0
            move_height = height + dy

        cdef Uint8 *srcptr = <Uint8 *> self.surface.pixels
        cdef Uint8 *destptr = <Uint8 *> self.surface.pixels

        srcptr += srcy * self.surface.pitch + srcx * per_pixel
        destptr += desty * self.surface.pitch + destx * per_pixel

        self.lock()

        with nogil:
            move_pixels(
                srcptr,
                destptr,
                move_height,
                move_width * per_pixel,
                self.surface.pitch,
                self.surface.pitch)

        self.unlock()

    def set_colorkey(self, color, flags=0):

        if color is None:
            if SDL_SetColorKey(self.surface, SDL_FALSE, 0):
                raise error()

        else:
            if SDL_SetColorKey(self.surface, SDL_TRUE, map_color(self.surface, color)):
                raise error()

    def get_colorkey(self):
        cdef Uint32 key

        if SDL_GetColorKey(self.surface, &key):
            return None

        return get_color(key, self.surface)

    def set_alpha(self, value, flags=0):
        if value is None:
            value = 255
            self.has_alpha = False
        else:
            self.has_alpha = True

        if SDL_SetSurfaceAlphaMod(self.surface, value):
            raise error()

    def get_alpha(self):
        cdef Uint8 rv

        if self.has_alpha or self.surface.format.Amask:

            if SDL_GetSurfaceAlphaMod(self.surface, &rv):
                raise error()

            return rv

        else:
            return None

    def lock(self, lock=None):
        cdef Surface root = self

        while root.parent:
            root = root.parent

        if lock is None:
            lock = self

        if root.locklist is None:
            root.locklist = [ ]

        root.locklist.append(lock)

        if SDL_LockSurface(root.surface):
            raise error()

    def unlock(self, lock=None):
        cdef Surface root = self

        while root.parent:
            root = root.parent

        if lock is None:
            lock = self

        if root.locklist is None:
            root.locklist = [ ]

        root.locklist.remove(lock)

        SDL_UnlockSurface(root.surface)

    def mustlock(self):
        cdef Surface root = self

        while root.parent:
            root = root.parent

        return SDL_MUSTLOCK(root.surface)

    def get_locked(self):
        if self.locklist:
            return True

    def get_locks(self):
        cdef Surface root = self

        while root.parent:
            root = root.parent

        if root.locklist is None:
            root.locklist = [ ]

        return root.locklist

    def get_at(self, pos):
        cdef int x, y
        cdef Uint8 *p

        x, y = pos

        if not (0 <= x < self.surface.w) or not (0 <= y < self.surface.h):
            raise IndexError("Position outside surface.")

        if self.surface.format.BytesPerPixel != 4:
            raise error("Surface has unsupported bytesize.")

        self.lock()

        p = <Uint8 *> self.surface.pixels
        p += y * self.surface.pitch
        p += x * 4

        cdef Uint32 pixel = (<Uint32 *> p)[0]

        self.unlock()

        return get_color(pixel, self.surface)

    def set_at(self, pos, color):
        cdef int x, y
        cdef Uint8 *p
        cdef Uint32 pixel

        x, y = pos

        if not (0 <= x < self.surface.w) or not (0 <= y < self.surface.h):
            raise ValueError("Position outside surface.")

        if self.surface.format.BytesPerPixel != 4:
            raise error("Surface has unsupported bytesize.")

        pixel = map_color(self.surface, color)

        self.lock()

        p = <Uint8 *> self.surface.pixels
        p += y * self.surface.pitch
        p += x * 4

        (<Uint32 *> p)[0] = pixel

        self.unlock()

    def get_at_mapped(self, pos):
        cdef int x, y
        cdef Uint8 *p

        x, y = pos

        if not (0 <= x < self.surface.w) or not (0 <= y < self.surface.h):
            raise ValueError("Position outside surface.")

        if self.surface.format.BytesPerPixel != 4:
            raise error("Surface has unsupported bytesize.")

        self.lock()

        p = <Uint8 *> self.surface.pixels
        p += y * self.surface.pitch
        p += x * 4

        cdef Uint32 pixel = (<Uint32 *> p)[0]

        self.unlock()

        return pixel

    def map_rgb(self, color):
        return map_color(self.surface, color)

    def unmap_rgb(self, pixel):
        return get_color(pixel, self.surface)

    def set_clip(self, rect):
        cdef SDL_Rect sdl_rect

        if rect is None:
            SDL_SetClipRect(self.surface, NULL)
        else:
            to_sdl_rect(rect, &sdl_rect)
            SDL_SetClipRect(self.surface, &sdl_rect)

    def get_clip(self):
        cdef SDL_Rect sdl_rect

        SDL_GetClipRect(self.surface, &sdl_rect)

        return (sdl_rect.x, sdl_rect.y, sdl_rect.w, sdl_rect.h)

    def subsurface(self, *args):
        cdef SDL_Rect sdl_rect

        if len(args) == 1:
            to_sdl_rect(args[0], &sdl_rect)
        else:
            to_sdl_rect(args, &sdl_rect)

        if sdl_rect.w < 0 or sdl_rect.h < 0:
            raise error("subsurface size must be non-negative.")

        if ((sdl_rect.x < 0)
            or (sdl_rect.y < 0)
            or (sdl_rect.x + sdl_rect.w > self.surface.w)
            or (sdl_rect.y + sdl_rect.h > self.surface.h)):

            raise error("subsurface rectangle outside surface area.")

        cdef Uint8 *pixels = <Uint8 *> self.surface.pixels
        pixels += sdl_rect.y * self.surface.pitch
        pixels += sdl_rect.x * self.surface.format.BytesPerPixel

        cdef SDL_Surface *new_surface = SDL_CreateRGBSurfaceFrom(
            pixels,
            sdl_rect.w,
            sdl_rect.h,
            self.surface.format.BitsPerPixel,
            self.surface.pitch,
            self.surface.format.Rmask,
            self.surface.format.Gmask,
            self.surface.format.Bmask,
            self.surface.format.Amask)

        if not new_surface:
            raise error()

        cdef Surface rv = Surface(())

        rv.surface = new_surface
        rv.parent = self
        rv.offset_x = sdl_rect.x
        rv.offset_y = sdl_rect.y

        if self.has_alpha:
            rv.set_alpha(self.get_alpha())

        return rv

    def get_parent(self):
        return self.parent

    def get_abs_parent(self):
        rv = self

        while rv.parent:
            rv = rv.parent

        return rv

    def get_offset(self):
        return (self.offset_x, self.offset_y)

    def get_abs_offset(self):
        cdef Surface surf = self

        cdef int offset_x = 0
        cdef int offset_y = 0

        while surf:
            offset_x += surf.offset_x
            offset_y += surf.offset_y
            surf = surf.parent

        return (offset_x, offset_y)

    def get_size(self):
        return self.surface.w, self.surface.h

    def get_width(self):
        return self.surface.w

    def get_height(self):
        return self.surface.h

    def get_rect(self, **kwargs):
        rv = Rect((0, 0, self.surface.w, self.surface.h))

        for k, v in kwargs.items():
            setattr(rv, k, v)

        return rv

    def get_bitsize(self):
        return self.surface.format.BitsPerPixel

    def get_bytesize(self):
        return self.surface.format.BytesPerPixel

    def get_flags(self):

        if self.get_window_flags:
            rv = self.get_window_flags()
        else:
            rv = 0

        if self.surface.format.Amask or self.has_alpha:
            rv = rv | SRCALPHA

        return rv

    def get_pitch(self):
        return self.surface.pitch

    def get_masks(self):
        cdef SDL_PixelFormat *format = self.surface.format
        return (format.Rmask, format.Gmask, format.Bmask, format.Amask)

    def set_masks(self, masks):
        warnings.warn("Surface.set_masks is not supported.")

    def get_shifts(self):
        cdef SDL_PixelFormat *format = self.surface.format
        return (format.Rshift, format.Gshift, format.Bshift, format.Ashift)

    def set_shifts(self, shifts):
        warnings.warn("Surface.set_shifts is not supported.")

    def get_shifts(self):
        cdef SDL_PixelFormat *format = self.surface.format
        return (format.Rshift, format.Gshift, format.Bshift, format.Ashift)

    def get_losses(self):
        cdef SDL_PixelFormat *format = self.surface.format
        return (format.Rloss, format.Gloss, format.Bloss, format.Aloss)

    def get_bounding_rect(self, min_alpha=1):

        cdef Uint32 amask = self.surface.format.Amask
        cdef Uint32 amin = (0x01010101 * min_alpha) & amask

        cdef int x
        cdef int y
        cdef int w
        cdef int h

        cdef int minx = self.surface.w - 1
        cdef int maxx = 0
        cdef int miny = self.surface.h - 1
        cdef int maxy = 0

        cdef Uint32 *row

        cdef Uint32 topleft
        cdef Uint32 botright

        if (not amask) or (self.surface.w == 0) or (self.surface.h == 0):
            return Rect((0, 0, self.surface.w, self.surface.h))

        self.lock()

        cdef Uint8 *pixels = <Uint8 *> self.surface.pixels

        with nogil:

            topleft = (<Uint32*> pixels)[0]
            botright = (<Uint32*> (pixels + self.surface.pitch * (self.surface.h - 1)))[self.surface.w - 1]

            if ((topleft & amask) > amin) and ((botright & amask) > amin):

                # Bounding box covers image.

                minx = 0
                miny = 0
                maxx = self.surface.w - 1
                maxy = self.surface.h - 1

            else:

                # Bounding box is smaller than image.

                for 0 <= y < self.surface.h:
                    row = <Uint32*> (pixels + self.surface.pitch * y)

                    for 0 <= x < self.surface.w:

                        if (row[x] & amask) >= amin:

                            if minx > x:
                                minx = x
                            if miny > y:
                                miny = y
                            if maxx < x:
                                maxx = x
                            if maxy < y:
                                maxy = y

        self.unlock()

        # Totally empty surface.
        if minx > maxx:
            return Rect((0, 0, 0, 0))

        x = minx
        y = miny
        w = min(maxx - minx + 1, self.surface.w - x)
        h = min(maxy - miny + 1, self.surface.h - y)

        return Rect((x, y, w, h))

    def get_view(self, kind='2'):
        raise error("Surface.get_view is not supported.")

    def get_buffer(self):
        cdef Uint8 *pixels = <Uint8 *> self.surface.pixels
        return pixels[self.surface.h * self.surface.pitch]

    property _pixels_address:
        def __get__(self):
            return <Uint64> self.surface.pixels


cdef api SDL_Surface *PySurface_AsSurface(surface):
    return (<Surface> surface).surface

cdef api object PySurface_New(SDL_Surface *surf):
    cdef Surface rv = Surface(())
    rv.take_surface(surf)
    return rv



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from sdl2_ttf cimport *
from pygame_sdl2.surface cimport *
from pygame_sdl2.color cimport *
from pygame_sdl2.rwobject cimport to_rwops

from pygame_sdl2.sysfont import SysFont, match_font, get_fonts
from pygame_sdl2.error import error
import pygame_sdl2

@pygame_sdl2.register_init
def init():
    TTF_Init()

@pygame_sdl2.register_quit
def quit(): # @ReservedAssignment
    TTF_Quit()

def get_init():
    return TTF_WasInit() != 0

def get_default_font():
    import os

    default = os.path.join(os.path.dirname(__file__), "DejaVuSans.ttf")

    if os.path.exists(default):
        return default

    return match_font("sans")

cdef class Font:
    cdef TTF_Font *font
    cdef int style

    def __cinit__(self):
        self.font = NULL
        self.style = TTF_STYLE_NORMAL

    def __dealloc__(self):
        if self.font:
            TTF_CloseFont(self.font)

    def __init__(self, fi, size):
        if fi is None:
            fi = get_default_font()
        self.font = TTF_OpenFontRW(to_rwops(fi), 1, size)
        if self.font == NULL:
            raise error()

    def render(self, text, antialias, color, background=None):
        cdef SDL_Surface *surf
        cdef SDL_Color fg

        if not text:
            w, h = self.size(" ")
            return Surface((0, h))

        to_sdl_color(color, &fg)

        TTF_SetFontStyle(self.font, self.style)
        if antialias:
            surf = TTF_RenderUTF8_Blended(self.font, text.encode('utf-8'), fg)
        else:
            surf = TTF_RenderUTF8_Solid(self.font, text.encode('utf-8'), fg)

        if surf == NULL:
            raise error()

        cdef Surface rv = Surface(())
        rv.take_surface(surf)

        if rv.surface.format.BitsPerPixel != 32:
            rv = rv.convert()

        if background is not None:
            bgsurf = rv.copy()
            bgsurf.fill(background)
            bgsurf.blit(rv, (0,0))
            return bgsurf
        return rv

    def size(self, text):
        cdef int w, h
        if TTF_SizeUTF8(self.font, text.encode('utf-8'), &w, &h) != 0:
            raise error()
        return w, h

    cdef set_style(self, flag, on):
        if on:
            self.style |= flag
        else:
            self.style &= ~flag

    def set_underline(self, on):
        self.set_style(TTF_STYLE_UNDERLINE, on)

    def get_underline(self):
        return self.style & TTF_STYLE_UNDERLINE

    def set_bold(self, on):
        self.set_style(TTF_STYLE_BOLD, on)

    def get_bold(self):
        return self.style & TTF_STYLE_BOLD

    def set_italic(self, on):
        self.set_style(TTF_STYLE_ITALIC, on)

    def get_italic(self):
        return self.style & TTF_STYLE_ITALIC

    def get_linesize(self):
        return TTF_FontLineSkip(self.font)

    def get_height(self):
        return TTF_FontHeight(self.font)

    def get_ascent(self):
        return TTF_FontAscent(self.font)

    def get_descent(self):
        return TTF_FontDescent(self.font)

    def metrics(self, text not None):
        cdef int minx, maxx, miny, maxy, advance
        cdef uint16_t chnum
        rv = []

        for ch in text:
            chnum = ord(ch)

            if TTF_GlyphMetrics(self.font, chnum, &minx, &maxx, &miny, &maxy, &advance) == 0:
                rv.append((minx, maxx, miny, maxy, advance))
            else:
                rv.append(None)

        return rv



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from pygame_sdl2.surface cimport Surface

import pygame_sdl2.gfxdraw as gfxdraw
from pygame_sdl2.rect import Rect
from pygame_sdl2.error import error

def rect(Surface surface, color, rect, width=0):
    if not isinstance(rect, Rect):
        rect = Rect(rect)

    if width == 0:
        gfxdraw.box(surface, rect, color)
    else:
        gfxdraw.rectangle(surface, rect, color)
        n = 1
        while n < width:
            r = Rect(rect.x - n, rect.y - n, rect.w + (n*2), rect.h + (n*2))
            gfxdraw.rectangle(surface, r, color)
            r = Rect(rect.x + n, rect.y + n, rect.w - (n*2), rect.h - (n*2))
            gfxdraw.rectangle(surface, r, color)
            n += 1
    dirty = Rect(rect.x - width, rect.y - width, rect.w + (width*2), rect.h + (width*2))
    return dirty.clip(surface.get_rect())

def polygon(Surface surface, color, pointlist, width=0):
    if width == 0:
        gfxdraw.filled_polygon(surface, pointlist, color)
        dirty = Rect(pointlist[0], (1, 1))
        n = 1
        while n < len(pointlist):
            dirty.union_ip(Rect(pointlist[n], (1,1)))
            n += 1
        return dirty.clip(surface.get_rect())
    else:
        return lines(surface, color, True, pointlist, width)

def circle(Surface surface, color, pos, radius, width=0):
    x, y = pos
    if width == 0:
        gfxdraw.filled_circle(surface, x, y, radius, color)
        dirty = Rect((x - radius, y - radius), (radius*2, radius*2))
        return dirty.clip(surface.get_rect())
    else:
        gfxdraw.circle(surface, x, y, radius, color)
        n = 1
        while n < width:
            gfxdraw.circle(surface, x, y, radius - n, color)
            gfxdraw.circle(surface, x + 1, y, radius - n, color)
            gfxdraw.circle(surface, x - 1, y, radius - n, color)
            n += 1
        dirty = Rect(x - radius - width, y - radius - width, (radius*2) + width, (radius*2) + width)
        return dirty.clip(surface.get_rect())

def ellipse(Surface surface, color, rect, width=0):
    raise error("Not implemented.")

def arc(Surface surface, color, rect, start_angle, stop_angle, width=1):
    raise error("Not implemented.")

def line(Surface surface, color, start_pos, end_pos, width=1):
    gfxdraw.thick_line(surface, start_pos[0], start_pos[1],
                       end_pos[0], end_pos[1], width, color)
    dirty = Rect(start_pos, (width, width))
    dirty.union_ip(Rect(end_pos, (width, width)))
    return dirty.clip(surface.get_rect())

def lines(Surface surface, color, closed, pointlist, width=1):
    n = 0
    dirty = Rect(pointlist[0], (width, width))
    while n < len(pointlist) - 1:
        line(surface, color, pointlist[n], pointlist[n+1], width)
        dirty.union_ip(Rect(pointlist[n+1], (width, width)))
        n += 1
    if closed:
        line(surface, color, pointlist[n], pointlist[0], width)
    return dirty.clip(surface.get_rect())

def aaline(Surface surface, color, startpos, endpos, blend=1):
    x1, y1 = startpos
    x2, y2 = endpos
    gfxdraw.aaline(surface, x1, y1, x2, y2, color)
    dirty = Rect(x1, y1, x2 - x1, y2 - y1)
    return dirty.clip(surface.get_rect())

def aalines(Surface surface, color, closed, pointlist, blend=1):
    n = 0
    dirty = Rect(pointlist[0], (1,1))
    while n < len(pointlist) - 1:
        r = aaline(surface, color, pointlist[n], pointlist[n+1])
        dirty.union_ip(r)
        n += 1
    if closed:
        aaline(surface, color, pointlist[n], pointlist[0])
    return dirty.clip(surface.get_rect())



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from sdl2_gfx cimport *
from libc.stdlib cimport malloc, free
from pygame_sdl2.surface cimport Surface
from pygame_sdl2.color cimport Color

from pygame_sdl2.error import error
from pygame_sdl2.rect import Rect


def pixel(Surface surface, x, y, color):
    cdef Color c = Color(color)
    pixelRGBA(surface.surface, x, y, c.r, c.g, c.b, c.a)

def hline(Surface surface, x1, x2, y, color):
    cdef Color c = Color(color)
    hlineRGBA(surface.surface, x1, x2, y, c.r, c.g, c.b, c.a)

def vline(Surface surface, x, y1, y2, color):
    cdef Color c = Color(color)
    vlineRGBA(surface.surface, x, y1, y2, c.r, c.g, c.b, c.a)

def rectangle(Surface surface, rect, color):
    cdef Color c = Color(color)
    if not isinstance(rect, Rect):
        rect = Rect(rect)
    rectangleRGBA(surface.surface, rect.x, rect.y, rect.x + rect.w, rect.y + rect.h, c.r, c.g, c.b, c.a)

def rounded_rectangle(Surface surface, rect, rad, color):
    cdef Color c = Color(color)
    if not isinstance(rect, Rect):
        rect = Rect(rect)
    roundedRectangleRGBA(surface.surface, rect.x, rect.y, rect.x + rect.w, rect.y + rect.h, rad, c.r, c.g, c.b, c.a)

def box(Surface surface, rect, color):
    cdef Color c = Color(color)
    if not isinstance(rect, Rect):
        rect = Rect(rect)
    boxRGBA(surface.surface, rect.x, rect.y, rect.x + rect.w, rect.y + rect.h, c.r, c.g, c.b, c.a)

def rounded_box(Surface surface, rect, rad, color):
    cdef Color c = Color(color)
    if not isinstance(rect, Rect):
        rect = Rect(rect)
    roundedBoxRGBA(surface.surface, rect.x, rect.y, rect.x + rect.w, rect.y + rect.h, rad, c.r, c.g, c.b, c.a)

def line(Surface surface, x1, y1, x2, y2, color):
    cdef Color c = Color(color)
    lineRGBA(surface.surface, x1, y1, x2, y2, c.r, c.g, c.b, c.a)

def aaline(Surface surface, x1, y1, x2, y2, color):
    cdef Color c = Color(color)
    aalineRGBA(surface.surface, x1, y1, x2, y2, c.r, c.g, c.b, c.a)

def thick_line(Surface surface, x1, y1, x2, y2, width, color):
    cdef Color c = Color(color)

    # This locks up in c code when trying to draw a zero-length line. So make
    # sure that doesn't happen.
    cdef int x1int, y1int, x2int, y2int
    x1int = x1
    y1int = y1
    x2int = x2
    y2int = y2


    if x1int == x2int and y1int == y2int:
        return

    thickLineRGBA(surface.surface, x1int, y1int, x2int, y2int, width, c.r, c.g, c.b, c.a)

def circle(Surface surface, x, y, r, color):
    cdef Color c = Color(color)
    circleRGBA(surface.surface, x, y, r, c.r, c.g, c.b, c.a)

def arc(Surface surface, x, y, r, start, end, color):
    cdef Color c = Color(color)
    arcRGBA(surface.surface, x, y, r, start, end, c.r, c.g, c.b, c.a)

def aacircle(Surface surface, x, y, r, color):
    cdef Color c = Color(color)
    aacircleRGBA(surface.surface, x, y, r, c.r, c.g, c.b, c.a)

def filled_circle(Surface surface, x, y, r, color):
    cdef Color c = Color(color)
    filledCircleRGBA(surface.surface, x, y, r, c.r, c.g, c.b, c.a)

def ellipse(Surface surface, x, y, rx, ry, color):
    cdef Color c = Color(color)
    ellipseRGBA(surface.surface, x, y, rx, ry, c.r, c.g, c.b, c.a)

def aaellipse(Surface surface, x, y, rx, ry, color):
    cdef Color c = Color(color)
    aaellipseRGBA(surface.surface, x, y, rx, ry, c.r, c.g, c.b, c.a)

def filled_ellipse(Surface surface, x, y, rx, ry, color):
    cdef Color c = Color(color)
    filledEllipseRGBA(surface.surface, x, y, rx, ry, c.r, c.g, c.b, c.a)

def pie(Surface surface, x, y, r, start, end, color):
    cdef Color c = Color(color)
    pieRGBA(surface.surface, x, y, r, start, end, c.r, c.g, c.b, c.a)

def filled_pie(Surface surface, x, y, r, start, end, color):
    cdef Color c = Color(color)
    filledPieRGBA(surface.surface, x, y, r, start, end, c.r, c.g, c.b, c.a)

def trigon(Surface surface, x1, y1, x2, y2, x3, y3, color):
    cdef Color c = Color(color)
    trigonRGBA(surface.surface, x1, y1, x2, y2, x3, y3, c.r, c.g, c.b, c.a)

def aatrigon(Surface surface, x1, y1, x2, y2, x3, y3, color):
    cdef Color c = Color(color)
    aatrigonRGBA(surface.surface, x1, y1, x2, y2, x3, y3, c.r, c.g, c.b, c.a)

def filled_trigon(Surface surface, x1, y1, x2, y2, x3, y3, color):
    cdef Color c = Color(color)
    filledTrigonRGBA(surface.surface, x1, y1, x2, y2, x3, y3, c.r, c.g, c.b, c.a)

def polygon(Surface surface, points, color):
    cdef Color c = Color(color)
    cdef Sint16 *vx
    cdef Sint16 *vy
    cdef size_t num_points = len(points)
    vx = <Sint16*>malloc(num_points * sizeof(Sint16))
    vy = <Sint16*>malloc(num_points * sizeof(Sint16))
    for n, pt in zip(range(num_points), points):
        vx[n], vy[n] = points[n]
    polygonRGBA(surface.surface, vx, vy, num_points, c.r, c.g, c.b, c.a)
    free(vx)
    free(vy)

def aapolygon(Surface surface, points, color):
    cdef Color c = Color(color)
    cdef Sint16 *vx
    cdef Sint16 *vy
    cdef size_t num_points = len(points)
    vx = <Sint16*>malloc(num_points * sizeof(Sint16))
    vy = <Sint16*>malloc(num_points * sizeof(Sint16))
    for n, pt in zip(range(num_points), points):
        vx[n], vy[n] = points[n]
    aapolygonRGBA(surface.surface, vx, vy, num_points, c.r, c.g, c.b, c.a)
    free(vx)
    free(vy)

def filled_polygon(Surface surface, points, color):
    cdef Color c = Color(color)
    cdef Sint16 *vx
    cdef Sint16 *vy
    cdef size_t num_points = len(points)
    vx = <Sint16*>malloc(num_points * sizeof(Sint16))
    vy = <Sint16*>malloc(num_points * sizeof(Sint16))
    for n, pt in zip(range(num_points), points):
        vx[n], vy[n] = points[n]
    filledPolygonRGBA(surface.surface, vx, vy, num_points, c.r, c.g, c.b, c.a)
    free(vx)
    free(vy)

def textured_polygon(Surface surface, points, Surface texture not None, tx, ty):
    cdef Sint16 *vx
    cdef Sint16 *vy
    cdef size_t num_points = len(points)
    vx = <Sint16*>malloc(num_points * sizeof(Sint16))
    vy = <Sint16*>malloc(num_points * sizeof(Sint16))
    for n, pt in zip(range(num_points), points):
        vx[n], vy[n] = points[n]
    texturedPolygon(surface.surface, vx, vy, num_points, texture.surface, tx, ty)
    free(vx)
    free(vy)

def bezier(Surface surface, points, steps, color):
    cdef Color c = Color(color)
    cdef Sint16 *vx
    cdef Sint16 *vy
    cdef size_t num_points = len(points)
    vx = <Sint16*>malloc(num_points * sizeof(Sint16))
    vy = <Sint16*>malloc(num_points * sizeof(Sint16))
    for n, pt in zip(range(num_points), points):
        vx[n], vy[n] = points[n]
    bezierRGBA(surface.surface, vx, vy, num_points, steps, c.r, c.g, c.b, c.a)
    free(vx)
    free(vy)



from sdl2 cimport *

cdef extern from "SDL2_rotozoom.h" nogil:
    cdef enum:
        SMOOTHING_OFF
        SMOOTHING_ON

    SDL_Surface *rotozoomSurface(SDL_Surface * src, double angle, double zoom, int smooth)
    SDL_Surface *rotozoomSurfaceXY(SDL_Surface * src, double angle, double zoomx, double zoomy, int smooth)
    void rotozoomSurfaceSize(int width, int height, double angle, double zoom, int *dstwidth, int *dstheight)
    void rotozoomSurfaceSizeXY(int width, int height, double angle, double zoomx, double zoomy, int *dstwidth, int *dstheight)

    SDL_Surface *zoomSurface(SDL_Surface * src, double zoomx, double zoomy, int smooth)
    void zoomSurfaceSize(int width, int height, double zoomx, double zoomy, int *dstwidth, int *dstheight)

    SDL_Surface *shrinkSurface(SDL_Surface * src, int factorx, int factory)
    SDL_Surface* rotateSurface90Degrees(SDL_Surface* src, int numClockwiseTurns)


cdef extern from "SDL_gfxPrimitives.h" nogil:
    int pixelColor(SDL_Surface * dst, Sint16 x, Sint16 y, Uint32 color)
    int pixelRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int hlineColor(SDL_Surface * dst, Sint16 x1, Sint16 x2, Sint16 y, Uint32 color)
    int hlineRGBA(SDL_Surface * dst, Sint16 x1, Sint16 x2, Sint16 y, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int vlineColor(SDL_Surface * dst, Sint16 x, Sint16 y1, Sint16 y2, Uint32 color)
    int vlineRGBA(SDL_Surface * dst, Sint16 x, Sint16 y1, Sint16 y2, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int rectangleColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint32 color)
    int rectangleRGBA(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int roundedRectangleColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 rad, Uint32 color)
    int roundedRectangleRGBA(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 rad, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int boxColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint32 color)
    int boxRGBA(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int roundedBoxColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 rad, Uint32 color)
    int roundedBoxRGBA(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 rad, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int lineColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint32 color)
    int lineRGBA(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int aalineColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint32 color)
    int aalineRGBA(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int thickLineColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 width, Uint32 color)
    int thickLineRGBA(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 width, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int circleColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Uint32 color)
    int circleRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int arcColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Sint16 start, Sint16 end, Uint32 color)
    int arcRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Sint16 start, Sint16 end, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int aacircleColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Uint32 color)
    int aacircleRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int filledCircleColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 r, Uint32 color)
    int filledCircleRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int ellipseColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rx, Sint16 ry, Uint32 color)
    int ellipseRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rx, Sint16 ry, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int aaellipseColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rx, Sint16 ry, Uint32 color)
    int aaellipseRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rx, Sint16 ry, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int filledEllipseColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rx, Sint16 ry, Uint32 color)
    int filledEllipseRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rx, Sint16 ry, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int pieColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Sint16 start, Sint16 end, Uint32 color)
    int pieRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Sint16 start, Sint16 end, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int filledPieColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Sint16 start, Sint16 end, Uint32 color)
    int filledPieRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 rad, Sint16 start, Sint16 end, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int trigonColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 x3, Sint16 y3, Uint32 color)
    int trigonRGBA(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 x3, Sint16 y3, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int aatrigonColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 x3, Sint16 y3, Uint32 color)
    int aatrigonRGBA(SDL_Surface * dst,  Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 x3, Sint16 y3, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int filledTrigonColor(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 x3, Sint16 y3, Uint32 color)
    int filledTrigonRGBA(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Sint16 x3, Sint16 y3, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int polygonColor(SDL_Surface * dst, const Sint16 * vx, const Sint16 * vy, int n, Uint32 color)
    int polygonRGBA(SDL_Surface * dst, const Sint16 * vx, const Sint16 * vy, int n, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int aapolygonColor(SDL_Surface * dst, const Sint16 * vx, const Sint16 * vy, int n, Uint32 color)
    int aapolygonRGBA(SDL_Surface * dst, const Sint16 * vx, const Sint16 * vy, int n, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int filledPolygonColor(SDL_Surface * dst, const Sint16 * vx, const Sint16 * vy, int n, Uint32 color)
    int filledPolygonRGBA(SDL_Surface * dst, const Sint16 * vx, const Sint16 * vy, int n, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    int texturedPolygon(SDL_Surface * dst, const Sint16 * vx, const Sint16 * vy, int n, SDL_Surface * texture,int texture_dx,int texture_dy)

    int bezierColor(SDL_Surface * dst, const Sint16 * vx, const Sint16 * vy, int n, int s, Uint32 color)
    int bezierRGBA(SDL_Surface * dst, const Sint16 * vx, const Sint16 * vy, int n, int s, Uint8 r, Uint8 g, Uint8 b, Uint8 a)

    void gfxPrimitivesSetFont(const void *fontdata, Uint32 cw, Uint32 ch)
    void gfxPrimitivesSetFontRotation(Uint32 rotation)
    int characterColor(SDL_Surface * dst, Sint16 x, Sint16 y, char c, Uint32 color)
    int characterRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, char c, Uint8 r, Uint8 g, Uint8 b, Uint8 a)
    int stringColor(SDL_Surface * dst, Sint16 x, Sint16 y, const char *s, Uint32 color)
    int stringRGBA(SDL_Surface * dst, Sint16 x, Sint16 y, const char *s, Uint8 r, Uint8 g, Uint8 b, Uint8 a)



from sdl2 cimport *

# Acquire the GIL for callbacks.
cdef extern from "SDL_mixer.h":
    void Mix_ChannelFinished(void (*channel_finished)(int channel))
    void Mix_HookMusicFinished(void (*music_finished)())
    void Mix_SetPostMix(void (*mix_func)(void *udata, Uint8 *stream, int len), void *arg)
    void Mix_HookMusic(void (*mix_func)(void *udata, Uint8 *stream, int len), void *arg)

cdef extern from "SDL_mixer.h" nogil:
    cdef enum:
        SDL_MIXER_MAJOR_VERSION
        SDL_MIXER_MINOR_VERSION
        SDL_MIXER_PATCHLEVEL

    const SDL_version * Mix_Linked_Version()

    ctypedef enum MIX_InitFlags:
        MIX_INIT_FLAC
        MIX_INIT_MP3
        MIX_INIT_OGG

    int Mix_Init(int flags)
    void Mix_Quit()

    cdef enum:
        MIX_CHANNELS
        MIX_DEFAULT_FREQUENCY
        MIX_DEFAULT_FORMAT
        MIX_DEFAULT_CHANNELS
        MIX_MAX_VOLUME

    ctypedef struct Mix_Chunk:
        int allocated
        Uint8 *abuf
        Uint32 alen
        Uint8 volume

    ctypedef enum Mix_Fading:
        MIX_NO_FADING
        MIX_FADING_OUT
        MIX_FADING_IN

    ctypedef enum Mix_MusicType:
        MUS_NONE
        MUS_CMD
        MUS_WAV
        MUS_MOD
        MUS_MID
        MUS_OGG
        MUS_MP3
        MUS_MP3_MAD
        MUS_FLAC
        MUS_MODPLUG

    ctypedef struct _Mix_Music
    ctypedef _Mix_Music Mix_Music

    int Mix_OpenAudio(int frequency, Uint16 format, int channels, int chunksize)
    int Mix_AllocateChannels(int numchans)
    int Mix_QuerySpec(int *frequency,Uint16 *format,int *channels)
    Mix_Chunk * Mix_LoadWAV_RW(SDL_RWops *src, int freesrc)
    Mix_Music * Mix_LoadMUS(const char *file)
    Mix_Music * Mix_LoadMUS_RW(SDL_RWops *src, int freesrc)
    Mix_Music * Mix_LoadMUSType_RW(SDL_RWops *src, Mix_MusicType type, int freesrc)
    Mix_Chunk * Mix_QuickLoad_WAV(Uint8 *mem)
    Mix_Chunk * Mix_QuickLoad_RAW(Uint8 *mem, Uint32 len)
    void Mix_FreeChunk(Mix_Chunk *chunk)
    void Mix_FreeMusic(Mix_Music *music)

    int Mix_GetNumChunkDecoders()
    const char * Mix_GetChunkDecoder(int index)
    int Mix_GetNumMusicDecoders()
    const char * Mix_GetMusicDecoder(int index)
    Mix_MusicType Mix_GetMusicType(const Mix_Music *music)
    void * Mix_GetMusicHookData()

    cdef enum:
        MIX_CHANNEL_POST

    ctypedef void (*Mix_EffectFunc_t)(int chan, void *stream, int len, void *udata)
    ctypedef void (*Mix_EffectDone_t)(int chan, void *udata)

    int Mix_RegisterEffect(int chan, Mix_EffectFunc_t f, Mix_EffectDone_t d, void *arg)
    int Mix_UnregisterEffect(int channel, Mix_EffectFunc_t f)
    int Mix_UnregisterAllEffects(int channel)

    cdef enum:
        MIX_EFFECTSMAXSPEED

    int Mix_SetPanning(int channel, Uint8 left, Uint8 right)
    int Mix_SetPosition(int channel, Sint16 angle, Uint8 distance)
    int Mix_SetDistance(int channel, Uint8 distance)
    int Mix_SetReverseStereo(int channel, int flip)

    int Mix_ReserveChannels(int num)
    int Mix_GroupChannel(int which, int tag)
    int Mix_GroupChannels(int from_, int to, int tag)
    int Mix_GroupAvailable(int tag)
    int Mix_GroupCount(int tag)
    int Mix_GroupOldest(int tag)
    int Mix_GroupNewer(int tag)

    int Mix_PlayChannelTimed(int channel, Mix_Chunk *chunk, int loops, int ticks)
    int Mix_PlayMusic(Mix_Music *music, int loops)

    int Mix_FadeInMusic(Mix_Music *music, int loops, int ms)
    int Mix_FadeInMusicPos(Mix_Music *music, int loops, int ms, double position)
    int Mix_FadeInChannelTimed(int channel, Mix_Chunk *chunk, int loops, int ms, int ticks)

    int Mix_Volume(int channel, int volume)
    int Mix_VolumeChunk(Mix_Chunk *chunk, int volume)
    int Mix_VolumeMusic(int volume)

    int Mix_HaltChannel(int channel)
    int Mix_HaltGroup(int tag)
    int Mix_HaltMusic()

    int Mix_ExpireChannel(int channel, int ticks)

    int Mix_FadeOutChannel(int which, int ms)
    int Mix_FadeOutGroup(int tag, int ms)
    int Mix_FadeOutMusic(int ms)

    Mix_Fading Mix_FadingMusic()
    Mix_Fading Mix_FadingChannel(int which)

    void Mix_Pause(int channel)
    void Mix_Resume(int channel)
    int Mix_Paused(int channel)

    void Mix_PauseMusic()
    void Mix_ResumeMusic()
    void Mix_RewindMusic()
    int Mix_PausedMusic()

    int Mix_SetMusicPosition(double position)

    int Mix_Playing(int channel)
    int Mix_PlayingMusic()

    int Mix_SetMusicCMD(const char *command)

    int Mix_SetSynchroValue(int value)
    int Mix_GetSynchroValue()

    int Mix_SetSoundFonts(const char *paths)
    const char* Mix_GetSoundFonts()
    int Mix_EachSoundFont(int (*function)(const char*, void*), void *data)

    Mix_Chunk * Mix_GetChunk(int channel)

    void Mix_CloseAudio()



# Copyright 2014 Patrick Dawson
# Copyright 2014 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.


from sdl2 cimport SDL_Surface

cdef class Surface:
    # Allow weak references.
    cdef object __weakref__

    # The SDL surface that corresponds to this surface.
    cdef SDL_Surface* surface

    # True if we own our surface. False if some other Surface owns our
    # surface.
    cdef bint owns_surface

    # True if this surface is owned by a window.
    cdef bint window_surface

    # A list of locks involving the surface.
    cdef object locklist

    # If this surface is a subsurface, the surface this surface is a subsurface
    # of.
    cdef Surface parent

    # If this surface has no parent, self. Otherwise, self.parent.root. The
    # ultimate parent.
    cdef Surface root

    # The offset of this surface within its parent.
    cdef int offset_x
    cdef int offset_y

    # If not None, a function that returns the window flags.
    cdef object get_window_flags

    cdef void take_surface(self, SDL_Surface *)

    cdef bint has_alpha



# Copyright 2014 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from __future__ import division, absolute_import, print_function

import os
import sys
import re
import subprocess
import platform

try:
    import setuptools
except:
    import distutils.core as setuptools

# The include and library dirs that we compile against.
include_dirs = [ ".", "src" ]
library_dirs = [ ]

# Extra arguments that will be given to the compiler.
extra_compile_args = [ ]
extra_link_args = [ ]

# Data files (including DLLs) to include with the package.
package_data = [ ]

# A list of extension objects that we use.
extensions = [ ]

# A list of macros that are defined for all modules.
global_macros = [ ]

# True if we're building on android.
android = "PYGAME_SDL2_ANDROID" in os.environ

# True if we're building on ios.
ios = "PYGAME_SDL2_IOS" in os.environ

# True if we're doing a static build.
static = "PYGAME_SDL2_STATIC" in os.environ

windows = platform.win32_ver()[0]

# The cython command.
if windows:
    cython_command = os.path.join(os.path.dirname(sys.executable), "Scripts", "cython.exe")
else:
    cython_command = "cython"

if sys.version_info[0] >= 3:
    version_flag = "-3"
    gen = "gen3"
else:
    version_flag = "-2"
    gen = "gen"

if static:
    gen = gen + "-static"


def system_path(path):
    """
    On windows/msys, converts a unix-style path returned from sdl2-config
    to a windows path. Otherwise, returns path.
    """

    if "MSYSTEM" in os.environ:
        path = subprocess.check_output([ "sh", "-c", "cmd //c echo " + path ]).strip()

    return path


def parse_cflags(command):
    """
    Runs `command`, and uses the command's output to set up include_dirs
    and extra_compile_args. if `command` is None, parses the cflags from
    the environment.
    """

    if "PYGAME_SDL2_CFLAGS" in os.environ:
        output = os.environ["PYGAME_SDL2_CFLAGS"]
    elif command is not None:
        output = subprocess.check_output(command, universal_newlines=True)
    else:
        output = os.environ.get("CFLAGS", "")

    for i in output.split():
        if i.startswith("-I"):
            include_dirs.append(system_path(i[2:]))
        else:
            extra_compile_args.append(i)


parse_cflags(None)


def parse_libs(command):
    """
    Runs `command`, and uses the command's output to set up library_dirs and
    extra_link_args. Returns a list of libraries to link against. If `command`
    is None, parses LDFLAGS from the environment.
    """

    if "PYGAME_SDL2_LDFLAGS" in os.environ:
        output = os.environ["PYGAME_SDL2_LDFLAGS"]
    else:
        output = subprocess.check_output(command, universal_newlines=True)

    libs = [ ]

    for i in output.split():
        if i.startswith("-L"):
            library_dirs.append(system_path(i[2:]))
        elif i.startswith("-l"):
            libs.append(i[2:])
        else:
            extra_compile_args.append(i)

    return libs


# A list of modules we do not wish to include.
exclude = set(os.environ.get("PYGAME_SDL2_EXCLUDE", "").split())


def cmodule(name, source, libs=[], define_macros=[]):
    """
    Compiles the python module `name` from the files given in
    `source`, and the libraries in `libs`.
    """

    if name in exclude:
        return

    extensions.append(setuptools.Extension(
        name,
        source,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        libraries=libs,
        define_macros=define_macros + global_macros,
        ))


necessary_gen = [ ]


def cython(name, source=[], libs=[], compile_if=True, define_macros=[]):
    """
    Compiles a cython module. This takes care of regenerating it as necessary
    when it, or any of the files it depends on, changes.
    """

    # Find the pyx file.
    split_name = name.split(".")

    fn = "src/" + "/".join(split_name) + ".pyx"

    if not os.path.exists(fn):
        print("Could not find {0}.".format(fn))
        sys.exit(-1)

    module_dir = os.path.dirname(fn)

    # Figure out what it depends on.
    deps = [ fn ]

    f = open(fn, "r")
    for l in f:

        m = re.search(r'from\s*([\w.]+)\s*cimport', l)
        if m:
            deps.append(m.group(1).replace(".", "/") + ".pxd")
            continue

        m = re.search(r'cimport\s*([\w.]+)', l)
        if m:
            deps.append(m.group(1).replace(".", "/") + ".pxd")
            continue

        m = re.search(r'include\s*"(.*?)"', l)
        if m:
            deps.append(m.group(1))
            continue
    f.close()

    # Filter out cython stdlib dependencies.
    deps = [ i for i in deps if (not i.startswith("cpython/")) and (not i.startswith("libc/")) ]

    # Determine if any of the dependencies are newer than the c file.
    c_fn = os.path.join(gen, name + ".c")
    necessary_gen.append(name + ".c")

    if os.path.exists(c_fn):
        c_mtime = os.path.getmtime(c_fn)
    else:
        c_mtime = 0

    out_of_date = False

    # print c_fn, "depends on", deps

    for dep_fn in deps:

        if os.path.exists(os.path.join("src", dep_fn)):
            dep_fn = os.path.join("src", dep_fn)
        elif os.path.exists(os.path.join("include", dep_fn)):
            dep_fn = os.path.join("include", dep_fn)
        elif os.path.exists(os.path.join(gen, dep_fn)):
            dep_fn = os.path.join(gen, dep_fn)
        elif os.path.exists(dep_fn):
            pass
        else:
            print("{0} depends on {1}, which can't be found.".format(fn, dep_fn))
            sys.exit(-1)

        if os.path.getmtime(dep_fn) > c_mtime:
            out_of_date = True

    if out_of_date and not cython_command:
        print("WARNING:", name, "is out of date, but RENPY_CYTHON isn't set.")
        out_of_date = False

    # If the file is out of date, regenerate it.
    if out_of_date:
        print(name, "is out of date.")

        try:
            subprocess.check_call([
                cython_command,
                version_flag,
                "-Iinclude",
                "-I" + gen,
                "-a",
                fn,
                "-o",
                c_fn])

            # Fix-up source for static loading
            if static:

                parent_module = '.'.join(split_name[:-1])
                parent_module_identifier = parent_module.replace('.', '_')

                with open(c_fn, 'r') as f:
                    ccode = f.read()

                with open(c_fn + ".dynamic", 'w') as f:
                    f.write(ccode)

                if len(split_name) > 1:

                    ccode = re.sub('Py_InitModule4\("([^"]+)"', 'Py_InitModule4("' + parent_module + '.\\1"', ccode) # Py2
                    ccode = re.sub('(__pyx_moduledef.*?"){}"'.format(re.escape(split_name[-1])), '\\1' + '.'.join(split_name) + '"', ccode, count=1, flags=re.DOTALL) # Py3
                    ccode = re.sub('^__Pyx_PyMODINIT_FUNC init', '__Pyx_PyMODINIT_FUNC init' + parent_module_identifier + '_', ccode, 0, re.MULTILINE) # Py2 Cython 0.28+
                    ccode = re.sub('^__Pyx_PyMODINIT_FUNC PyInit_', '__Pyx_PyMODINIT_FUNC PyInit_' + parent_module_identifier + '_', ccode, 0, re.MULTILINE) # Py3 Cython 0.28+
                    ccode = re.sub('^PyMODINIT_FUNC init', 'PyMODINIT_FUNC init' + parent_module_identifier + '_', ccode, 0, re.MULTILINE) # Py2 Cython 0.25.2

                cname = "_".join(split_name)

                ccode += """

static struct _inittab CNAME_inittab[] = {
#if PY_MAJOR_VERSION < 3
    { "PYNAME", initCNAME },
#else
    { "PYNAME", PyInit_CNAME },
#endif
    { NULL, NULL },
};

static void CNAME_constructor(void) __attribute__((constructor));

static void CNAME_constructor(void) {
    PyImport_ExtendInittab(CNAME_inittab);
}
""".replace("PYNAME", name).replace("CNAME", cname)

                with open(c_fn, 'w') as f:
                    f.write(ccode)

        except subprocess.CalledProcessError as e:
            print()
            print(str(e))
            print()
            sys.exit(-1)

    # Build the module normally once we have the c file.
    if compile_if:
        cmodule(name, [ c_fn ] + source, libs=libs, define_macros=define_macros)


def find_unnecessary_gen():

    for i in os.listdir(gen):
        if not i.endswith(".c"):
            continue

        if i in necessary_gen:
            continue

        print("Unnecessary file", os.path.join(gen, i))


py_modules = [ ]


def pymodule(name):
    """
    Causes a python module to be included in the build.
    """

    if name in exclude:
        return

    py_modules.append(name)


def setup(name, version, **kwargs):
    """
    Calls the distutils setup function.
    """

    global extensions

    if (len(sys.argv) >= 2) and (sys.argv[1] == "generate"):
        return

    if "--no-extensions" in sys.argv:
        sys.argv = [ i for i in sys.argv if i != "--no-extensions" ]
        extensions = [ ]

    setuptools.setup(
        name=name,
        version=version,
        ext_modules=extensions,
        py_modules=py_modules,
        packages=[ name ],
        package_dir={ name : 'src/' + name },
        package_data={ name : package_data },
        zip_safe=False,
        **kwargs
        )


# Start in the directory containing setup.py.
os.chdir(os.path.abspath(os.path.dirname(sys.argv[0])))

# Ensure the gen directory exists.
if not os.path.exists(gen):
    os.mkdir(gen)



from sdl2 cimport *

cdef extern from "SDL_image.h" nogil:
    ctypedef enum IMG_InitFlags:
        IMG_INIT_JPG
        IMG_INIT_PNG
        IMG_INIT_TIF
        IMG_INIT_WEBP

    int IMG_Init(int flags)
    void IMG_Quit()

    SDL_Surface *IMG_Load(const char *file)
    SDL_Surface *IMG_Load_RW(SDL_RWops *src, int freesrc)
    SDL_Surface *IMG_LoadTyped_RW(SDL_RWops *src, int freesrc, const char *type)
    
    SDL_Texture *IMG_LoadTexture(SDL_Renderer *renderer, const char *file)
    SDL_Texture *IMG_LoadTexture_RW(SDL_Renderer *renderer, SDL_RWops *src, int freesrc)
    SDL_Texture *IMG_LoadTextureTyped_RW(SDL_Renderer *renderer, SDL_RWops *src, int freesrc, const char *type)

    int IMG_SavePNG(SDL_Surface *surface, const char *file)
    int IMG_SavePNG_RW(SDL_Surface *surface, SDL_RWops *dst, int freedst)



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from sdl2_image cimport *
from libc.string cimport memcpy, memset
from pygame_sdl2.display cimport *
from pygame_sdl2.surface cimport *
from pygame_sdl2.rwobject cimport to_rwops
from pygame_sdl2.rect cimport to_sdl_rect

from pygame_sdl2.rect import Rect
from pygame_sdl2.error import error
from pygame_sdl2.color import Color
import json
import warnings

BLENDMODE_NONE = SDL_BLENDMODE_NONE
BLENDMODE_BLEND = SDL_BLENDMODE_BLEND
BLENDMODE_ADD = SDL_BLENDMODE_ADD
BLENDMODE_MOD = SDL_BLENDMODE_MOD

cdef bint DEBUG_DRAW_BBOX = True

cdef rinfo_to_dict(SDL_RendererInfo *rinfo):
    # Ignore texture_formats for now.
    return {
            "name" : rinfo.name,
            "software" : rinfo.flags & SDL_RENDERER_SOFTWARE != 0,
            "accelerated" : rinfo.flags & SDL_RENDERER_ACCELERATED != 0,
            "vsync" : rinfo.flags & SDL_RENDERER_PRESENTVSYNC != 0,
            "rtt" : rinfo.flags & SDL_RENDERER_TARGETTEXTURE != 0,
            "max_texture_width" : rinfo.max_texture_width,
            "max_texture_height" : rinfo.max_texture_height,
    }

def get_drivers():
    cdef SDL_RendererInfo rinfo
    cdef int num_drivers
    rv = []

    num_drivers = SDL_GetNumRenderDrivers()
    for n in range(num_drivers):
        if SDL_GetRenderDriverInfo(n, &rinfo) != 0:
            raise error()

        rv.append(rinfo_to_dict(&rinfo))

    return rv


cdef class Renderer:
    cdef SDL_Renderer *renderer
    cdef dict _info

    def __cinit__(self):
        self.renderer = NULL

    def __dealloc__(self):
        if self.renderer:
            SDL_DestroyRenderer(self.renderer)

    def __init__(self, Window window=None, vsync=False, driver=-1):
        if window is None:
            window = main_window

        cdef uint32_t flags = SDL_RENDERER_ACCELERATED
        if vsync:
            flags |= SDL_RENDERER_PRESENTVSYNC

        self.renderer = SDL_CreateRenderer(window.window, driver, flags)
        if self.renderer == NULL:
            self.renderer = SDL_GetRenderer(window.window)
            if self.renderer == NULL:
                raise error()

        cdef SDL_RendererInfo rinfo
        if SDL_GetRendererInfo(self.renderer, &rinfo) != 0:
            raise error()

        self._info = rinfo_to_dict(&rinfo)

        if not self.info()["accelerated"]:
            warnings.warn("Renderer is not accelerated.")

    def load_texture(self, fi):
        cdef SDL_Texture *tex
        cdef Texture t = Texture()

        if isinstance(fi, Surface):
            tex = SDL_CreateTextureFromSurface(self.renderer, (<Surface>fi).surface)
        else:
            tex = IMG_LoadTexture_RW(self.renderer, to_rwops(fi), 1)

        if tex == NULL:
            raise error()

        t.set(self.renderer, tex)
        return TextureNode(t)

    def load_atlas(self, filename):
        """ Loads a file in the popular JSON (Hash) format exported by
            TexturePacker and other software. """

        return TextureAtlas(self, filename)

    def render_present(self):
        with nogil:
            SDL_RenderPresent(self.renderer)

    def info(self):
        return self._info

    cdef set_drawcolor(self, col):
        if not isinstance(col, Color):
            col = Color(col)
        SDL_SetRenderDrawColor(self.renderer, col.r, col.g, col.b, col.a)

    def clear(self, color):
        self.set_drawcolor(color)
        SDL_RenderClear(self.renderer)

    def draw_line(self, color not None, x1, y1, x2, y2):
        self.set_drawcolor(color)
        if SDL_RenderDrawLine(self.renderer, x1, y1, x2, y2) != 0:
            raise error()

    def draw_point(self, color not None, x, y):
        self.set_drawcolor(color)
        SDL_RenderDrawPoint(self.renderer, x, y)

    def draw_rect(self, color not None, rect):
        cdef SDL_Rect r
        to_sdl_rect(rect, &r)
        self.set_drawcolor(color)
        SDL_RenderDrawRect(self.renderer, &r)

    def fill_rect(self, color not None, rect):
        cdef SDL_Rect r
        to_sdl_rect(rect, &r)
        self.set_drawcolor(color)
        SDL_RenderFillRect(self.renderer, &r)

    def set_viewport(self, rect=None):
        cdef SDL_Rect vprect
        if rect is None:
            SDL_RenderSetViewport(self.renderer, NULL)
        else:
            to_sdl_rect(rect, &vprect)
            SDL_RenderSetViewport(self.renderer, &vprect)

    def create_texture(self, size):
        if SDL_RenderTargetSupported(self.renderer) != SDL_TRUE:
            raise error()


cdef class Texture:
    """ Mostly for internal use. Users should only see this for RTT. """

    cdef SDL_Renderer *renderer
    cdef SDL_Texture *texture
    cdef public int w, h

    def __cinit__(self):
        self.texture = NULL

    def __dealloc__(self):
        if self.texture:
            SDL_DestroyTexture(self.texture)

    cdef set(self, SDL_Renderer *ren, SDL_Texture *tex):
        cdef Uint32 format
        cdef int access, w, h

        self.renderer = ren
        self.texture = tex

        if SDL_QueryTexture(self.texture, &format, &access, &w, &h) != 0:
            raise error()

        self.w = w
        self.h = h


cdef class TextureNode:
    """ A specified area of a texture. """

    cdef Texture texture

    # The absolute rect within the texture.
    cdef SDL_Rect source_rect

    # The relative rect within the original tile.
    cdef SDL_Rect trimmed_rect

    # The dimensions of the original tile.
    cdef int source_w
    cdef int source_h

    def __init__(self, tex):
        if isinstance(tex, Texture):
            self.texture = tex
            to_sdl_rect((0,0,tex.w,tex.h), &self.source_rect)
            to_sdl_rect((0,0,tex.w,tex.h), &self.trimmed_rect)
            self.source_w = tex.w
            self.source_h = tex.h

        elif isinstance(tex, TextureNode):
            self.texture = (<TextureNode>tex).texture

        else:
            raise ValueError()

    def render(self, dest=None):
        cdef SDL_Rect dest_rect

        if dest is None:
            with nogil:
                SDL_RenderCopy(self.texture.renderer, self.texture.texture, NULL, NULL)

        else:
            to_sdl_rect(dest, &dest_rect)
            with nogil:
                if dest_rect.w == 0 or dest_rect.h == 0:
                    dest_rect.w = self.trimmed_rect.w
                    dest_rect.h = self.trimmed_rect.h
                SDL_RenderCopy(self.texture.renderer, self.texture.texture, NULL, &dest_rect)


cdef class TextureAtlas:
    cdef object frames

    def __init__(self, Renderer ren, fi):
        jdata = json.load(open(fi, "r"))
        image = jdata["meta"]["image"]

        cdef TextureNode tn = ren.load_texture(image)

        self.frames = {}

        cdef TextureNode itex
        for itm in jdata["frames"].items():
            iname, idict = itm
            itex = TextureNode(tn)
            f = idict["frame"]
            to_sdl_rect((f['x'], f['y'], f['w'], f['h']), &itex.source_rect, "frame")
            f = idict["spriteSourceSize"]
            to_sdl_rect((f['x'], f['y'], f['w'], f['h']), &itex.trimmed_rect, "spriteSourceSize")
            if idict["rotated"]:
                raise error("Rotation not supported yet.")
            itex.source_w = idict["sourceSize"]["w"]
            itex.source_h = idict["sourceSize"]["h"]

            self.frames[iname] = itex

    def __getitem__(self, key):
        return self.frames[key]

    def keys(self):
        return self.frames.keys()


cdef class Sprite:
    """ One or more TextureNodes, with possible transforms applied. """

    cdef list nodes
    cdef SDL_Rect _pos
    cdef SDL_Rect bounding_box

    cdef double _rotation
    cdef int _flip
    cdef SDL_Color _color
    cdef double _scalex
    cdef double _scaley

    def __init__(self, nodes):
        self._color.r = 255
        self._color.g = 255
        self._color.b = 255
        self._color.a = 255
        self._scalex = 1.0
        self._scaley = 1.0
        self._flip = SDL_FLIP_NONE

        memset(&self._pos, 0, sizeof(SDL_Rect))
        memset(&self.bounding_box, 0, sizeof(SDL_Rect))

        if isinstance(nodes, TextureNode):
            nodes = [nodes]

        self.nodes = []
        # TODO: Check that they're all from the same texture.
        for node in nodes:
            if not isinstance(node, TextureNode):
                raise ValueError("Invalid argument: %s" % node)
            SDL_UnionRect(&self.bounding_box, &(<TextureNode>node).trimmed_rect,
                &self.bounding_box)
            self.nodes.append(node)

    cdef void adjust_rect(Sprite self, const SDL_Rect *dest, const SDL_Rect *rin, SDL_Rect *rout) nogil:
        rout.x = dest.x + <int>(self._scalex * rin.x)
        rout.y = dest.y + <int>(self._scaley * rin.y)
        rout.w = <int>(self._scalex * rin.w)
        rout.h = <int>(self._scaley * rin.h)

    def render(self, dest=None):
        cdef Texture tex = (<TextureNode>self.nodes[0]).texture
        cdef SDL_Rect dest_rect
        cdef SDL_Rect real_dest
        cdef SDL_Point pivot

        if dest is None:
            memcpy(&dest_rect, &self._pos, sizeof(SDL_Rect))
        else:
            to_sdl_rect(dest, &dest_rect)

        with nogil:
            SDL_SetTextureColorMod(tex.texture, self._color.r, self._color.g, self._color.b)
            SDL_SetTextureAlphaMod(tex.texture, self._color.a)

            if DEBUG_DRAW_BBOX:
                # TODO: Adjust for rotation.
                self.adjust_rect(&dest_rect, &self.bounding_box, &real_dest)

                SDL_SetRenderDrawColor(tex.renderer, 0x00, 0x00, 0x00, 0x00)

        cdef TextureNode tn
        for x in self.nodes:
            tn = <TextureNode> x

            with nogil:
                pivot.x = <int>(self._scalex * (tn.source_w / 2 - tn.trimmed_rect.x))
                pivot.y = <int>(self._scaley * (tn.source_h / 2 - tn.trimmed_rect.y))

                self.adjust_rect(&dest_rect, &tn.trimmed_rect, &real_dest)

                SDL_RenderCopyEx(tex.renderer, tex.texture, &tn.source_rect,
                    &real_dest, self._rotation, &pivot,
                    <SDL_RendererFlip>self._flip)

    def collides(self, Sprite other not None):
        cdef SDL_Rect r1, r2

        self.adjust_rect(&self._pos, &self.bounding_box, &r1)
        other.adjust_rect(&other._pos, &other.bounding_box, &r2)

        return SDL_HasIntersection(&r1, &r2) == SDL_TRUE

    property pos:
        def __get__(self):
            return self._pos.x, self._pos.y

        def __set__(self, val):
            self._pos.x = val[0]
            self._pos.y = val[1]

    property color:
        def __set__(self, val):
            if not isinstance(val, Color):
                val = Color(val)

            self._color.r = val.r
            self._color.g = val.g
            self._color.b = val.b
            self._color.a = val.a

    property alpha:
        def __get__(self):
            return self._color.a

        def __set__(self, val):
            self._color.a = val

    property rotation:
        def __get__(self):
            return self._rotation

        def __set__(self, val):
            self._rotation = val

    property scale:
        def __get__(self):
            if self._scalex == self._scaley:
                return self._scalex
            else:
                return self._scalex, self._scaley

        def __set__(self, arg):
            if type(arg) == tuple:
                x, y = arg
            else:
                x = y = arg

            self._scalex = x
            self._scaley = y

    property hflip:
        def __get__(self):
            return self._flip & SDL_FLIP_HORIZONTAL

        def __set__(self, val):
            if val:
                self._flip |= SDL_FLIP_HORIZONTAL
            else:
                self._flip &= ~SDL_FLIP_HORIZONTAL

    property vflip:
        def __get__(self):
            return self._flip & SDL_FLIP_VERTICAL

        def __set__(self, val):
            if val:
                self._flip |= SDL_FLIP_VERTICAL
            else:
                self._flip &= ~SDL_FLIP_VERTICAL


cdef class Container:
    """ Multiple sprites, positioned relative to the container. """

    cdef SDL_Rect _rect
    cdef list sprites

    def __init__(self, rect):
        """ Parameter may be a position (no clipping) or a rect (clipped). """

        self.sprites = []
        to_sdl_rect(rect, &self._rect)
        if len(rect) == 2:
            self._rect.w = self._rect.h = 0

    def add(self, Sprite sprite not None):
        self.sprites.append(sprite)

    def render(self, dest=None):
        # TODO: Something other than this to get the SDL_Renderer.
        cdef SDL_Renderer *ren = (<TextureNode>(<Sprite>self.sprites[0]).nodes[0]).texture.renderer

        if self._rect.w != 0 and self._rect.h != 0:
            SDL_RenderSetClipRect(ren, &self._rect)

        for s in self.sprites:
            s.render((s.pos[0] + self._rect.x, s.pos[1] + self._rect.y))

        # TODO: Save and restore previous clip rect instead?
        SDL_RenderSetClipRect(ren, NULL)

    property pos:
        def __get__(self):
            return self._rect.x, self._rect.y

        def __set__(self, val):
            self._rect.x = val[0]
            self._rect.y = val[1]

    property rect:
        def __get__(self):
            return Rect(self._rect.x, self._rect.y, self._rect.w, self._rect.h)

        def __set__(self, val):
            to_sdl_rect(val, &self._rect)




# Copyright 2014 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

# This file exists because this file must be python, and not cython. That's
# because the init function time.pyx is inittime, which conflicts with the
# built-in inittime function (used by the time module) when everything is
# statically linked, as it is on iOS.
from pygame_sdl2.pygame_time import *




# Automatically generated by generate_keycode_list.py
# Do not edit!

K_UNKNOWN = SDLK_UNKNOWN
K_RETURN = SDLK_RETURN
K_ESCAPE = SDLK_ESCAPE
K_BACKSPACE = SDLK_BACKSPACE
K_TAB = SDLK_TAB
K_SPACE = SDLK_SPACE
K_EXCLAIM = SDLK_EXCLAIM
K_QUOTEDBL = SDLK_QUOTEDBL
K_HASH = SDLK_HASH
K_PERCENT = SDLK_PERCENT
K_DOLLAR = SDLK_DOLLAR
K_AMPERSAND = SDLK_AMPERSAND
K_QUOTE = SDLK_QUOTE
K_LEFTPAREN = SDLK_LEFTPAREN
K_RIGHTPAREN = SDLK_RIGHTPAREN
K_ASTERISK = SDLK_ASTERISK
K_PLUS = SDLK_PLUS
K_COMMA = SDLK_COMMA
K_MINUS = SDLK_MINUS
K_PERIOD = SDLK_PERIOD
K_SLASH = SDLK_SLASH
K_0 = SDLK_0
K_1 = SDLK_1
K_2 = SDLK_2
K_3 = SDLK_3
K_4 = SDLK_4
K_5 = SDLK_5
K_6 = SDLK_6
K_7 = SDLK_7
K_8 = SDLK_8
K_9 = SDLK_9
K_COLON = SDLK_COLON
K_SEMICOLON = SDLK_SEMICOLON
K_LESS = SDLK_LESS
K_EQUALS = SDLK_EQUALS
K_GREATER = SDLK_GREATER
K_QUESTION = SDLK_QUESTION
K_AT = SDLK_AT
K_LEFTBRACKET = SDLK_LEFTBRACKET
K_BACKSLASH = SDLK_BACKSLASH
K_RIGHTBRACKET = SDLK_RIGHTBRACKET
K_CARET = SDLK_CARET
K_UNDERSCORE = SDLK_UNDERSCORE
K_BACKQUOTE = SDLK_BACKQUOTE
K_a = SDLK_a
K_b = SDLK_b
K_c = SDLK_c
K_d = SDLK_d
K_e = SDLK_e
K_f = SDLK_f
K_g = SDLK_g
K_h = SDLK_h
K_i = SDLK_i
K_j = SDLK_j
K_k = SDLK_k
K_l = SDLK_l
K_m = SDLK_m
K_n = SDLK_n
K_o = SDLK_o
K_p = SDLK_p
K_q = SDLK_q
K_r = SDLK_r
K_s = SDLK_s
K_t = SDLK_t
K_u = SDLK_u
K_v = SDLK_v
K_w = SDLK_w
K_x = SDLK_x
K_y = SDLK_y
K_z = SDLK_z
K_CAPSLOCK = SDLK_CAPSLOCK
K_F1 = SDLK_F1
K_F2 = SDLK_F2
K_F3 = SDLK_F3
K_F4 = SDLK_F4
K_F5 = SDLK_F5
K_F6 = SDLK_F6
K_F7 = SDLK_F7
K_F8 = SDLK_F8
K_F9 = SDLK_F9
K_F10 = SDLK_F10
K_F11 = SDLK_F11
K_F12 = SDLK_F12
K_PRINTSCREEN = SDLK_PRINTSCREEN
K_SCROLLLOCK = SDLK_SCROLLLOCK
K_PAUSE = SDLK_PAUSE
K_INSERT = SDLK_INSERT
K_HOME = SDLK_HOME
K_PAGEUP = SDLK_PAGEUP
K_DELETE = SDLK_DELETE
K_END = SDLK_END
K_PAGEDOWN = SDLK_PAGEDOWN
K_RIGHT = SDLK_RIGHT
K_LEFT = SDLK_LEFT
K_DOWN = SDLK_DOWN
K_UP = SDLK_UP
K_NUMLOCKCLEAR = SDLK_NUMLOCKCLEAR
K_KP_DIVIDE = SDLK_KP_DIVIDE
K_KP_MULTIPLY = SDLK_KP_MULTIPLY
K_KP_MINUS = SDLK_KP_MINUS
K_KP_PLUS = SDLK_KP_PLUS
K_KP_ENTER = SDLK_KP_ENTER
K_KP_1 = SDLK_KP_1
K_KP_2 = SDLK_KP_2
K_KP_3 = SDLK_KP_3
K_KP_4 = SDLK_KP_4
K_KP_5 = SDLK_KP_5
K_KP_6 = SDLK_KP_6
K_KP_7 = SDLK_KP_7
K_KP_8 = SDLK_KP_8
K_KP_9 = SDLK_KP_9
K_KP_0 = SDLK_KP_0
K_KP_PERIOD = SDLK_KP_PERIOD
K_APPLICATION = SDLK_APPLICATION
K_POWER = SDLK_POWER
K_KP_EQUALS = SDLK_KP_EQUALS
K_F13 = SDLK_F13
K_F14 = SDLK_F14
K_F15 = SDLK_F15
K_F16 = SDLK_F16
K_F17 = SDLK_F17
K_F18 = SDLK_F18
K_F19 = SDLK_F19
K_F20 = SDLK_F20
K_F21 = SDLK_F21
K_F22 = SDLK_F22
K_F23 = SDLK_F23
K_F24 = SDLK_F24
K_EXECUTE = SDLK_EXECUTE
K_HELP = SDLK_HELP
K_MENU = SDLK_MENU
K_SELECT = SDLK_SELECT
K_STOP = SDLK_STOP
K_AGAIN = SDLK_AGAIN
K_UNDO = SDLK_UNDO
K_CUT = SDLK_CUT
K_COPY = SDLK_COPY
K_PASTE = SDLK_PASTE
K_FIND = SDLK_FIND
K_MUTE = SDLK_MUTE
K_VOLUMEUP = SDLK_VOLUMEUP
K_VOLUMEDOWN = SDLK_VOLUMEDOWN
K_KP_COMMA = SDLK_KP_COMMA
K_KP_EQUALSAS400 = SDLK_KP_EQUALSAS400
K_ALTERASE = SDLK_ALTERASE
K_SYSREQ = SDLK_SYSREQ
K_CANCEL = SDLK_CANCEL
K_CLEAR = SDLK_CLEAR
K_PRIOR = SDLK_PRIOR
K_RETURN2 = SDLK_RETURN2
K_SEPARATOR = SDLK_SEPARATOR
K_OUT = SDLK_OUT
K_OPER = SDLK_OPER
K_CLEARAGAIN = SDLK_CLEARAGAIN
K_CRSEL = SDLK_CRSEL
K_EXSEL = SDLK_EXSEL
K_KP_00 = SDLK_KP_00
K_KP_000 = SDLK_KP_000
K_THOUSANDSSEPARATOR = SDLK_THOUSANDSSEPARATOR
K_DECIMALSEPARATOR = SDLK_DECIMALSEPARATOR
K_CURRENCYUNIT = SDLK_CURRENCYUNIT
K_CURRENCYSUBUNIT = SDLK_CURRENCYSUBUNIT
K_KP_LEFTPAREN = SDLK_KP_LEFTPAREN
K_KP_RIGHTPAREN = SDLK_KP_RIGHTPAREN
K_KP_LEFTBRACE = SDLK_KP_LEFTBRACE
K_KP_RIGHTBRACE = SDLK_KP_RIGHTBRACE
K_KP_TAB = SDLK_KP_TAB
K_KP_BACKSPACE = SDLK_KP_BACKSPACE
K_KP_A = SDLK_KP_A
K_KP_B = SDLK_KP_B
K_KP_C = SDLK_KP_C
K_KP_D = SDLK_KP_D
K_KP_E = SDLK_KP_E
K_KP_F = SDLK_KP_F
K_KP_XOR = SDLK_KP_XOR
K_KP_POWER = SDLK_KP_POWER
K_KP_PERCENT = SDLK_KP_PERCENT
K_KP_LESS = SDLK_KP_LESS
K_KP_GREATER = SDLK_KP_GREATER
K_KP_AMPERSAND = SDLK_KP_AMPERSAND
K_KP_DBLAMPERSAND = SDLK_KP_DBLAMPERSAND
K_KP_VERTICALBAR = SDLK_KP_VERTICALBAR
K_KP_DBLVERTICALBAR = SDLK_KP_DBLVERTICALBAR
K_KP_COLON = SDLK_KP_COLON
K_KP_HASH = SDLK_KP_HASH
K_KP_SPACE = SDLK_KP_SPACE
K_KP_AT = SDLK_KP_AT
K_KP_EXCLAM = SDLK_KP_EXCLAM
K_KP_MEMSTORE = SDLK_KP_MEMSTORE
K_KP_MEMRECALL = SDLK_KP_MEMRECALL
K_KP_MEMCLEAR = SDLK_KP_MEMCLEAR
K_KP_MEMADD = SDLK_KP_MEMADD
K_KP_MEMSUBTRACT = SDLK_KP_MEMSUBTRACT
K_KP_MEMMULTIPLY = SDLK_KP_MEMMULTIPLY
K_KP_MEMDIVIDE = SDLK_KP_MEMDIVIDE
K_KP_PLUSMINUS = SDLK_KP_PLUSMINUS
K_KP_CLEAR = SDLK_KP_CLEAR
K_KP_CLEARENTRY = SDLK_KP_CLEARENTRY
K_KP_BINARY = SDLK_KP_BINARY
K_KP_OCTAL = SDLK_KP_OCTAL
K_KP_DECIMAL = SDLK_KP_DECIMAL
K_KP_HEXADECIMAL = SDLK_KP_HEXADECIMAL
K_LCTRL = SDLK_LCTRL
K_LSHIFT = SDLK_LSHIFT
K_LALT = SDLK_LALT
K_LGUI = SDLK_LGUI
K_RCTRL = SDLK_RCTRL
K_RSHIFT = SDLK_RSHIFT
K_RALT = SDLK_RALT
K_RGUI = SDLK_RGUI
K_MODE = SDLK_MODE
K_AUDIONEXT = SDLK_AUDIONEXT
K_AUDIOPREV = SDLK_AUDIOPREV
K_AUDIOSTOP = SDLK_AUDIOSTOP
K_AUDIOPLAY = SDLK_AUDIOPLAY
K_AUDIOMUTE = SDLK_AUDIOMUTE
K_MEDIASELECT = SDLK_MEDIASELECT
K_WWW = SDLK_WWW
K_MAIL = SDLK_MAIL
K_CALCULATOR = SDLK_CALCULATOR
K_COMPUTER = SDLK_COMPUTER
K_AC_SEARCH = SDLK_AC_SEARCH
K_AC_HOME = SDLK_AC_HOME
K_AC_BACK = SDLK_AC_BACK
K_AC_FORWARD = SDLK_AC_FORWARD
K_AC_STOP = SDLK_AC_STOP
K_AC_REFRESH = SDLK_AC_REFRESH
K_AC_BOOKMARKS = SDLK_AC_BOOKMARKS
K_BRIGHTNESSDOWN = SDLK_BRIGHTNESSDOWN
K_BRIGHTNESSUP = SDLK_BRIGHTNESSUP
K_DISPLAYSWITCH = SDLK_DISPLAYSWITCH
K_KBDILLUMTOGGLE = SDLK_KBDILLUMTOGGLE
K_KBDILLUMDOWN = SDLK_KBDILLUMDOWN
K_KBDILLUMUP = SDLK_KBDILLUMUP
K_EJECT = SDLK_EJECT
K_SLEEP = SDLK_SLEEP
K_APP1 = SDLK_APP1
K_APP2 = SDLK_APP2
K_AUDIOREWIND = SDLK_AUDIOREWIND
K_AUDIOFASTFORWARD = SDLK_AUDIOFASTFORWARD



from sdl2 cimport *
cimport sdl2

from pygame_sdl2.rect import Rect
from pygame_sdl2.color import Color

include "event_list.pxi"
ACTIVEEVENT = SDL_LASTEVENT - 1
VIDEORESIZE = SDL_LASTEVENT - 2
VIDEOEXPOSE = SDL_LASTEVENT - 3

include "keycode_list.pxi"

include "glattr.pxi"
GL_SWAP_CONTROL = -1

GL_CONTEXT_PROFILE_CORE = SDL_GL_CONTEXT_PROFILE_CORE
GL_CONTEXT_PROFILE_COMPATIBILITY = SDL_GL_CONTEXT_PROFILE_COMPATIBILITY
GL_CONTEXT_PROFILE_ES = SDL_GL_CONTEXT_PROFILE_ES

RLEACCEL = SDL_RLEACCEL
SRCCOLORKEY = 0

# Surface flags.
SRCALPHA  = 0x80000000
HWSURFACE = 0x40000000
SWSURFACE = 0

# Window flags.
FULLSCREEN = SDL_WINDOW_FULLSCREEN
OPENGL = SDL_WINDOW_OPENGL
NOFRAME = SDL_WINDOW_BORDERLESS
RESIZABLE = SDL_WINDOW_RESIZABLE
DOUBLEBUF = 0

WINDOW_FULLSCREEN = SDL_WINDOW_FULLSCREEN
WINDOW_SHOWN = SDL_WINDOW_SHOWN
WINDOW_HIDDEN = SDL_WINDOW_HIDDEN
WINDOW_OPENGL = SDL_WINDOW_OPENGL
WINDOW_NOFRAME = SDL_WINDOW_BORDERLESS
WINDOW_BORDERLESS = SDL_WINDOW_BORDERLESS
WINDOW_RESIZABLE = SDL_WINDOW_RESIZABLE
WINDOW_MINIMIZED = SDL_WINDOW_MINIMIZED
WINDOW_MAXIMIZED = SDL_WINDOW_MAXIMIZED
WINDOW_INPUT_GRABBED = SDL_WINDOW_INPUT_GRABBED
WINDOW_INPUT_FOCUS = SDL_WINDOW_MOUSE_FOCUS
WINDOW_FULLSCREEN_DESKTOP = SDL_WINDOW_FULLSCREEN_DESKTOP
WINDOW_FOREIGN = SDL_WINDOW_FOREIGN
WINDOW_ALLOW_HIGHDPI = SDL_WINDOW_ALLOW_HIGHDPI

WINDOWPOS_CENTERED = SDL_WINDOWPOS_CENTERED
WINDOWPOS_UNDEFINED = SDL_WINDOWPOS_UNDEFINED

# Blend modes
BLEND_RGB_ADD = BLEND_ADD = 0x1
BLEND_RGB_SUB = BLEND_SUB = 0x2
BLEND_RGB_MULT = BLEND_MULT = 0x3
BLEND_RGB_MIN = BLEND_MIN = 0x4
BLEND_RGB_MAX = BLEND_MAX = 0x5

BLEND_RGBA_ADD = 0x6
BLEND_RGBA_SUB = 0x7
BLEND_RGBA_MULT = 0x8
BLEND_RGBA_MIN = 0x9
BLEND_RGBA_MAX = 0x10

BLEND_PREMULTIPLIED = 0x11

# Scrap types.
SCRAP_TEXT = "text/plain"

BYTEORDER = SDL_BYTEORDER
LIL_ENDIAN = SDL_LIL_ENDIAN
BIG_ENDIAN = SDL_BIG_ENDIAN

SDL_VERSION_TUPLE = (SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_PATCHLEVEL)

globals().update(dict(
    KMOD_NONE = sdl2.KMOD_NONE,

    KMOD_LSHIFT = sdl2.KMOD_LSHIFT,
    KMOD_RSHIFT = sdl2.KMOD_RSHIFT,
    KMOD_SHIFT = sdl2.KMOD_SHIFT,

    KMOD_LCTRL = sdl2.KMOD_LCTRL,
    KMOD_RCTRL = sdl2.KMOD_RCTRL,
    KMOD_CTRL = sdl2.KMOD_CTRL,

    KMOD_LALT = sdl2.KMOD_LALT,
    KMOD_RALT = sdl2.KMOD_RALT,
    KMOD_ALT = sdl2.KMOD_ALT,

    KMOD_LGUI = sdl2.KMOD_LGUI,
    KMOD_RGUI = sdl2.KMOD_RGUI,
    KMOD_GUI = sdl2.KMOD_GUI,

    KMOD_LMETA = sdl2.KMOD_LGUI,
    KMOD_RMETA = sdl2.KMOD_RGUI,
    KMOD_META = sdl2.KMOD_GUI,

    KMOD_NUM = sdl2.KMOD_NUM,
    KMOD_CAPS = sdl2.KMOD_CAPS,

    KMOD_MODE = sdl2.KMOD_MODE,

    K_FIRST = 0,
    K_LAST = sdl2.SDLK_SLEEP,

    K_BREAK = sdl2.SDLK_PAUSE,
    K_EURO = sdl2.SDLK_CURRENCYUNIT,

    K_KP0 = sdl2.SDLK_KP_0,
    K_KP1 = sdl2.SDLK_KP_1,
    K_KP2 = sdl2.SDLK_KP_2,
    K_KP3 = sdl2.SDLK_KP_3,
    K_KP4 = sdl2.SDLK_KP_4,
    K_KP5 = sdl2.SDLK_KP_5,
    K_KP6 = sdl2.SDLK_KP_6,
    K_KP7 = sdl2.SDLK_KP_7,
    K_KP8 = sdl2.SDLK_KP_8,
    K_KP9 = sdl2.SDLK_KP_9,

    K_LMETA = sdl2.SDLK_LGUI,
    K_LSUPER = sdl2.SDLK_LGUI,
    K_RMETA = sdl2.SDLK_RGUI,
    K_RSUPER = sdl2.SDLK_RGUI,

    K_NUMLOCK = sdl2.SDLK_NUMLOCKCLEAR,
    K_SCROLLOCK = sdl2.SDLK_SCROLLLOCK,
    K_PRINT = sdl2.SDLK_PRINTSCREEN,
))

TOUCH_MOUSEID = -1

include "controller.pxi"

POWERSTATE_UNKNOWN = SDL_POWERSTATE_UNKNOWN
POWERSTATE_ON_BATTERY = SDL_POWERSTATE_ON_BATTERY
POWERSTATE_NO_BATTERY = SDL_POWERSTATE_NO_BATTERY
POWERSTATE_CHARGING = SDL_POWERSTATE_CHARGING
POWERSTATE_CHARGED = SDL_POWERSTATE_CHARGED




# Copyright 2014 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from cpython.ref cimport Py_INCREF, Py_DECREF
from libc.string cimport memcpy
from libc.stdio cimport FILE, fopen, fclose, fseek, ftell, fread, SEEK_SET, SEEK_CUR, SEEK_END
from libc.stdlib cimport calloc, free

from pygame_sdl2.compat import file_type, bytes_, unicode_

import sys
import io

# The fsencoding.
fsencoding = sys.getfilesystemencoding() or "utf-8"

cdef extern from "SDL.h" nogil:
    Sint64 SDL_RWtell(SDL_RWops* context)

    Sint64 SDL_RWseek(SDL_RWops* context,
                      Sint64     offset,
                      int        whence)

    size_t SDL_RWread(SDL_RWops* context,
                  void*             ptr,
                  size_t            size,
                  size_t            maxnum)

    size_t SDL_RWwrite(SDL_RWops* context,
                       const void*       ptr,
                       size_t            size,
                       size_t            num)

    int SDL_RWclose(SDL_RWops* context)


cdef extern from "python_threads.h":
    void init_python_threads()


cdef set_error(e):
    cdef char *msg
    e = str(e)
    msg = <char *> e
    SDL_SetError("%s", msg)

cdef Sint64 python_size(SDL_RWops *context) with gil:
    f = <object> context.hidden.unknown.data1

    try:
        cur = f.tell()
        f.seek(0, 2)
        rv = f.tell()
        f.seek(cur, 0)
    except:
        return -1

    return rv

cdef Sint64 python_seek(SDL_RWops *context, Sint64 seek, int whence) with gil:
    f = <object> context.hidden.unknown.data1

    try:
        f.seek(seek, whence)
        rv = f.tell()
    except Exception as e:
        set_error(e)
        return -1

    return rv

cdef size_t python_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) with gil:
    f = <object> context.hidden.unknown.data1

    try:
        data = f.read(size * maxnum)
    except Exception as e:
        set_error(e)
        return -1

    memcpy(ptr, <void *><char *> data, len(data))
    return len(data)

cdef size_t python_write(SDL_RWops *context, const void *ptr, size_t size, size_t maxnum) with gil:
    f = <object> context.hidden.unknown.data1
    data = (<char *> ptr)[:size * maxnum]

    try:
        f.write(data)
    except Exception as e:
        set_error(e)
        return -1

    return len(data)

cdef int python_close(SDL_RWops *context) with gil:
    f = <object> context.hidden.unknown.data1

    try:
        f.close()
    except Exception as e:
        set_error(e)
        return -1

    Py_DECREF(f)
    SDL_FreeRW(context)
    return 0

cdef struct SubFile:
    SDL_RWops *rw
    Sint64 base
    Sint64 length
    Sint64 tell

cdef Sint64 subfile_size(SDL_RWops *context) nogil:
    cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1
    return sf.length

cdef Sint64 subfile_seek(SDL_RWops *context, Sint64 seek, int whence) nogil:
    cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1

    if whence == RW_SEEK_SET:
        sf.tell = SDL_RWseek(sf.rw, seek + sf.base, RW_SEEK_SET) - sf.base
    elif whence == RW_SEEK_CUR:
        sf.tell = SDL_RWseek(sf.rw, seek, RW_SEEK_CUR) - sf.base
    elif whence == RW_SEEK_END:
        sf.tell = SDL_RWseek(sf.rw, sf.base + sf.length + seek, RW_SEEK_SET) - sf.base

    return sf.tell

cdef size_t subfile_read(SDL_RWops *context, void *ptr, size_t size, size_t maxnum) nogil:
    cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1

    cdef Sint64 left = sf.length - sf.tell
    cdef size_t rv;

    if size * maxnum > left:
        maxnum = left // size

    if maxnum == 0:
        return 0

    rv = SDL_RWread(sf.rw, ptr, size, maxnum)

    if rv > 0:
        sf.tell += size * rv

    return rv

cdef int subfile_close(SDL_RWops *context) nogil:
    cdef SubFile *sf = <SubFile *> context.hidden.unknown.data1

    SDL_RWclose(sf.rw)
    free(sf)
    SDL_FreeRW(context)


cdef SDL_RWops *to_rwops(filelike, mode="rb") except NULL:

    cdef FILE *f
    cdef SubFile *sf
    cdef SDL_RWops *rv
    cdef SDL_RWops *rw
    cdef char *cname
    cdef char *cmode

    if not isinstance(mode, bytes_):
        mode = mode.encode("ascii")

    name = filelike

    if isinstance(filelike, (file_type, io.IOBase)) and mode == "rb":
        name = getattr(filelike, "name", None)

    # Try to open as a file.
    if isinstance(name, bytes_):
        name = name.decode(fsencoding)
    elif isinstance(name, unicode_):
        pass
    else:
        name = None

    if name is not None:

        dname = name.encode("utf-8")
        cname = dname
        cmode = mode

        with nogil:
            rv = SDL_RWFromFile(cname, cmode)

        if rv == NULL:
            raise IOError("Could not open {!r}: {}".format(name, SDL_GetError()))

        try:
            filelike.close()
        except:
            pass

        return rv

    if mode == b"rb":
        try:

            # If we have these properties, we're either an APK asset or a Ren'Py-style
            # subfile, so use an optimized path.
            name = filelike.name
            base = filelike.base
            length = filelike.length

            if name is not None:

                if not isinstance(name, unicode_):
                    name = name.decode(fsencoding)

                dname = name.encode("utf-8")
                cname = dname

                with nogil:
                    rw = SDL_RWFromFile(cname, b"rb")

                if not rw:
                    raise IOError("Could not open {!r}.".format(name))

                SDL_RWseek(rw, base, RW_SEEK_SET);

                sf = <SubFile *> calloc(sizeof(SubFile), 1)
                sf.rw = rw
                sf.base = base
                sf.length = length
                sf.tell = 0;

                rv = SDL_AllocRW()
                rv.size = subfile_size
                rv.seek = subfile_seek
                rv.read = subfile_read
                rv.write = NULL
                rv.close = subfile_close
                rv.type = 0
                rv.hidden.unknown.data1 = <void *> sf

                try:
                    filelike.close()
                except:
                    pass

                return rv

        except AttributeError:
            pass


    if not (hasattr(filelike, "read") or hasattr(filelike, "write")):
        raise IOError("{!r} is not a filename or file-like object.".format(filelike))

    Py_INCREF(filelike)

    rv = SDL_AllocRW()
    rv.size = python_size
    rv.seek = python_seek
    rv.read = python_read
    rv.write = python_write
    rv.close = python_close
    rv.type = 0
    rv.hidden.unknown.data1 = <void *> filelike

    return rv


cdef api SDL_RWops *RWopsFromPython(filelike) except NULL:
    return to_rwops(filelike)

init_python_threads()



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy
from pygame_sdl2.display cimport Window, main_window
from pygame_sdl2.rect cimport to_sdl_rect

from pygame_sdl2.error import error


cdef class KeyboardState:
    # Allow weak references.
    cdef object __weakref__

    cdef uint8_t *data
    cdef int numkeys

    def __cinit__(self):
        self.data = NULL

    def __dealloc__(self):
        if self.data != NULL:
            free(self.data)

    def __init__(self):
        cdef uint8_t *state = SDL_GetKeyboardState(&self.numkeys)
        if state == NULL:
            raise error()

        # Take a snapshot of the current state, rather than saving the pointer.
        self.data = <uint8_t*>malloc(self.numkeys)
        memcpy(self.data, state, self.numkeys)

    def __getitem__(self, int key):
        if SDLK_DELETE < key < SDLK_CAPSLOCK:
            raise IndexError("Out of range.")

        cdef int sc = <int>SDL_GetScancodeFromKey(key)
        if sc > self.numkeys:
            raise IndexError("Out of range.")

        return self.data[sc]


def get_focused():
    return SDL_GetKeyboardFocus() != NULL

def get_pressed():
    """ No longer returns a tuple. Use the returned object to check for
        individual keys, but don't loop through it. """

    return KeyboardState()

def get_mods():
    return SDL_GetModState()

def set_mods(state):
    SDL_SetModState(state)

def set_repeat(delay=0, interval=0):
    # Not possible with SDL2.
    pass

def get_repeat():
    # Not possible with SDL2.
    return (0,0)

def name(key):
    return SDL_GetKeyName(key)

text_input = False

def start_text_input():
    global text_input
    text_input = True

    if SDL_HasScreenKeyboardSupport():
        SDL_StartTextInput()

def stop_text_input():
    global text_input
    text_input = False

    if SDL_HasScreenKeyboardSupport():
        SDL_StopTextInput()

def set_text_input_rect(rect):
    cdef SDL_Rect sdl_rect;

    if rect is not None:
        to_sdl_rect(rect, &sdl_rect)
        SDL_SetTextInputRect(&sdl_rect)
    else:
        SDL_SetTextInputRect(NULL)

def has_screen_keyboard_support():
    return SDL_HasScreenKeyboardSupport()

def is_screen_keyboard_shown(Window window=None):
    if window is None:
        window = main_window

    return SDL_IsScreenKeyboardShown(window.window)



from __future__ import unicode_literals, print_function

import os
import argparse
import sys
import shutil
import sysconfig

def main():
    venv = os.environ.get("VIRTUAL_ENV", None)
    if not venv:
        print("VIRTUAL_ENV is not set.")
        sys.exit(1)

    for dn in os.listdir(os.path.join(venv, "include")):
        if dn.startswith("python"):
            break

    target = os.path.join(venv, "include", dn)

    try:
        source = os.readlink(target)
    except:
        print(target, "is not a symlink. Perhaps this script has already been run.")
        sys.exit(1)

    tmp = target + ".tmp"

    if os.path.exists(tmp):
        shutil.rmtree(tmp)

    os.mkdir(tmp)

    for i in os.listdir(source):
        if i == "pygame_sdl2":
            continue

        os.symlink(os.path.join(source, i), os.path.join(tmp, i))

    os.unlink(target)
    os.rename(tmp, target)

if __name__ == "__main__":
    main()



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import cython
import math
from sdl2 cimport *
from pygame_sdl2.error import error
import pygame_sdl2

cdef int timer_id = 0

@pygame_sdl2.register_init
def init():
    pygame_sdl2.display.sdl_main_init()

    if SDL_InitSubSystem(SDL_INIT_TIMER):
        raise error()

@pygame_sdl2.register_quit
def quit(): # @ReservedAssignment
    SDL_QuitSubSystem(SDL_INIT_TIMER)

def get_ticks():
    return SDL_GetTicks()

def wait(int milliseconds):
    cdef int start

    start = SDL_GetTicks()
    with nogil:
        SDL_Delay(milliseconds)
    return SDL_GetTicks() - start

def delay(milliseconds):
    # SDL_Delay() should be accurate enough.
    return wait(milliseconds)

cdef Uint32 timer_callback(Uint32 interval, void *param) nogil:
    cdef SDL_Event e
    e.type = <int>param
    e.user.code = 0
    e.user.data1 = NULL
    e.user.data2 = NULL
    SDL_PushEvent(&e)
    return interval

# A map from eventid to SDL_Timer_ID.
cdef dict timer_by_event = { }

def set_timer(eventid, milliseconds):

    cdef int timer_id = timer_by_event.get(eventid, 0)

    if timer_id != 0:
        SDL_RemoveTimer(timer_id)
        timer_id = 0

    if milliseconds > 0:
        timer_id = SDL_AddTimer(milliseconds, <SDL_TimerCallback>timer_callback, <void*><int>eventid)
        if timer_id == 0:
            raise error()

    timer_by_event[eventid] = timer_id

class Clock:
    def __init__(self):
        self.last = SDL_GetTicks()
        self.last_frames = []
        self.frametime = 0
        self.raw_frametime = 0

    def tick(self, framerate=0):
        cdef int now = SDL_GetTicks()
        self.raw_frametime = now - self.last
        while len(self.last_frames) > 9:
            self.last_frames.pop(0)
        if framerate == 0:
            self.last = now
            self.last_frames.append(self.raw_frametime)
            return self.raw_frametime
        cdef int frame_duration = 1.0 / framerate * 1000
        if self.raw_frametime < frame_duration:
            delay(frame_duration - self.raw_frametime)
        now = SDL_GetTicks()
        self.frametime = now - self.last
        self.last = now
        self.last_frames.append(self.frametime)
        return self.frametime

    def tick_busy_loop(self, framerate=0):
        return self.tick(framerate)

    def get_time(self):
        return self.frametime

    def get_rawtime(self):
        return self.raw_frametime

    @cython.cdivision(True)
    def get_fps(self):
        cdef int total_time = sum(self.last_frames)
        cdef float average_time = total_time / 1000.0 / len(self.last_frames)
        cdef float average_fps = 1.0 / average_time
        return 0 if math.isnan(average_fps) else average_fps



# Copyright 2014 Tom Rothamel <tom@rothamel.us>
# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from __future__ import division, print_function, absolute_import

import sys
import importlib


class MissingModule(object):

    def __init__(self, name, reason):
        self.__name__ = name
        self.reason = reason

    def __getattr__(self, attr):
        raise NotImplementedError(self.reason)


def try_import(name):
    full_name = "pygame_sdl2." + name

    try:
        module = importlib.import_module(full_name)
    except (IOError, ImportError) as e:
        module = MissingModule(full_name, "Could not import {}: {}".format(full_name, str(e)))

    globals()[name] = module
    sys.modules[full_name] = module


# Lists of functions that are called on init and quit.
init_functions = [ ]
quit_functions = [ ]


def register_init(fn):
    init_functions.append(fn)
    return fn


def register_quit(fn):
    quit_functions.append(fn)
    return fn


def init():
    numpass = 0
    numfail = 0

    for i in init_functions:
        try:
            i()
            numpass += 1
        except:
            numfail += 1

    return numpass, numfail


def quit():  # @ReservedAssignment
    for i in quit_functions:
        try:
            i()
        except:
            pass


# Import core modules.
from pygame_sdl2.error import *

from pygame_sdl2.surface import Surface
from pygame_sdl2.rect import Rect

import pygame_sdl2.color
import pygame_sdl2.display
import pygame_sdl2.event
import pygame_sdl2.key
import pygame_sdl2.locals  # @ReservedAssignment
import pygame_sdl2.time
import pygame_sdl2.version
import pygame_sdl2.locals as constants

# Import optional modules.
try_import("controller")
try_import("draw")
try_import("font")
try_import("image")
try_import("joystick")
try_import("mixer")
try_import("mouse")
try_import("power")
try_import("transform")
try_import("scrap")
try_import("sprite")
try_import("sysfont")

# Optional imports should be included in this function, so they show up
# in packaging tools (like py2exe).


def _optional_imports():
    import pygame_sdl2.compat
    import pygame_sdl2.controller
    import pygame_sdl2.rwobject
    import pygame_sdl2.gfxdraw
    import pygame_sdl2.draw
    import pygame_sdl2.font
    import pygame_sdl2.image
    import pygame_sdl2.joystick
    import pygame_sdl2.mixer
    import pygame_sdl2.mouse
    import pygame_sdl2.power
    import pygame_sdl2.transform
    import pygame_sdl2.scrap
    import pygame_sdl2.sprite
    import pygame_sdl2.sysfont


# Fill this module with locals.
from pygame_sdl2.locals import *


def import_as_pygame():
    """
    Imports pygame_sdl2 as pygame, so that running the 'import pygame.whatever'
    statement will import pygame_sdl2.whatever instead.
    """

    import os
    import warnings

    if "PYGAME_SDL2_USE_PYGAME" in os.environ:
        return

    if "pygame" in sys.modules:
        warnings.warn("Pygame has already been imported, import_as_pygame may not work.", stacklevel=2)

    for name, mod in list(sys.modules.items()):
        name = name.split('.')
        if name[0] != 'pygame_sdl2':
            continue

        name[0] = 'pygame'
        name = ".".join(name)

        sys.modules[name] = mod

    sys.modules['pygame.constants'] = constants


def get_sdl_byteorder():
    return BYTEORDER


def get_sdl_version():
    return SDL_VERSION_TUPLE


get_platform = pygame_sdl2.display.get_platform

# We have new-style buffers, but not the pygame.newbuffer module.
HAVE_NEWBUF = False



# Copyright 2015 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from pygame_sdl2.error import error
import pygame_sdl2

from pygame_sdl2.rwobject cimport to_rwops

@pygame_sdl2.register_init
def init():
    """
    Initializes game controller support.
    """

    pygame_sdl2.display.sdl_main_init()

    if SDL_InitSubSystem(SDL_INIT_GAMECONTROLLER):
        raise error()

@pygame_sdl2.register_quit
def quit(): # @ReservedAssignment
    """
    Shuts down game controller support.
    """

    SDL_QuitSubSystem(SDL_INIT_GAMECONTROLLER)

def get_init():
    """
    Returns true if game controller support has been initialized.
    """

    return SDL_WasInit(SDL_INIT_GAMECONTROLLER) != 0

def get_count():
    """
    Returns the number of joysticks and game controllers connected to the
    system. This is one more than the maximum index that can be passed to
    Controller. While every game controller is a joystick, not every joystick
    is a game controller. To check that index `i` corresponds to a game
    controller, use code such as::

        if pygame_sdl2.controller.Controller(i).is_controller():
            ...
    """

    return SDL_NumJoysticks()


def add_mapping(mapping):
    """
    Adds a game controller mapping from the string in `mapping`.
    """

    if SDL_GameControllerAddMapping(mapping) == -1:
        raise error()

def add_mappings(mapping_file):
    """
    Adds game controller mappings from `mapping_file`, which can be a string
    filename or a file-like object. This file should be gamecontrollerdb.txt,
    downloaded from:

    https://raw.githubusercontent.com/gabomdq/SDL_GameControllerDB/master/gamecontrollerdb.txt

    This should be called before creating a Controller object. It can be called
    multiple times to add multiple database files.
    """

    cdef SDL_RWops *rwops = to_rwops(mapping_file)

    if SDL_GameControllerAddMappingsFromRW(rwops, 1) == -1:
        raise error()

def get_axis_from_string(name):
    """
    Returns the axis number of the controller axis with `name`, or
    pygame.CONTROLLER_AXIS_INVALID if `name` is not known.
    """

    return SDL_GameControllerGetAxisFromString(name)

def get_button_from_string(name):
    """
    Returns the button number of the controller button with `name`, or
    pygame.CONTROLLER_BUTTON_INVALID if `name` is not known.
    """

    return SDL_GameControllerGetButtonFromString(name)

def get_string_for_axis(axis):
    """
    Returns a string describing the controller axis `axis`, which must be
    an integer. Returns None if the axis is not known.
    """

    cdef const char *rv = SDL_GameControllerGetStringForAxis(axis)

    if rv != NULL:
        return rv
    else:
        return None

def get_string_for_button(button):
    """
    Returns a string describing the controller button `button`, which must be
    an integer. Returns None if the button is not known.
    """

    cdef const char *rv = SDL_GameControllerGetStringForButton(button)

    if rv != NULL:
        return rv
    else:
        return None


cdef class Controller:
    # Allow weak references.
    cdef object __weakref__

    cdef SDL_GameController *controller
    cdef int index

    def __cinit__(self):
        self.controller = NULL

    def __init__(self, index):
        """
        Represents a game controller object corresponding to the joystick
        with `index`.
        """

        self.index = index

    def init(self):
        """
        Opens the game controller, causing it to begin sending events.
        """

        if self.controller == NULL:
            self.controller = SDL_GameControllerOpen(self.index)
            if self.controller == NULL:
                raise error()

    def quit(self): # @ReservedAssignment
        """
        Closes the game controller, preventing it from sending events.
        """

        if self.controller and SDL_GameControllerGetAttached(self.controller):
            SDL_GameControllerClose(self.controller)
            self.controller = NULL

    def get_init(self):
        """
        Returns true if the controller has been initialized, false otherwise.
        """

        if self.controller:
            return True
        else:
            return False

    def get_axis(self, axis):
        """
        Returns the value of `axis`, which is one of the pygame.CONTROLLER_AXIS
        constants.

        Axes return values from -32768 to 32678, triggers return 0 to 32768. 0 is
        returned on failure.
        """

        if self.controller == NULL:
            raise error("controller not initialized.")

        return SDL_GameControllerGetAxis(self.controller, axis)


    def get_button(self, button):
        """
        Returns the value of `button`, which must be one of the pygame.CONTROLLER_BUTTON
        constants.

        Returns 1 if the button is pressed, 0 if not or the button does not exist.
        """

        if self.controller == NULL:
            raise error("controller not initialized.")

        return SDL_GameControllerGetButton(self.controller, button)

    def get_name(self):
        """
        Returns an implementation-dependent name for this game controller,
        or None if no name could be found.
        """

        cdef const char *rv = SDL_GameControllerNameForIndex(self.index)

        if rv == NULL:
            return None

        return rv

    def is_controller(self):
        """
        Returns true if this Controller object corresponds to a supported
        game controller, or False otherwise.
        """

        return SDL_IsGameController(self.index)

    def get_guid_string(self):
        """
        Returns the guid string corresponding to this controller.
        """

        cdef SDL_JoystickGUID guid
        cdef char s[33]

        guid = SDL_JoystickGetDeviceGUID(self.index)
        SDL_JoystickGetGUIDString(guid, s, 33)

        return s




# Automatically generated by generate_event_list.py
# Do not edit!

cdef extern from "SDL.h":
    enum SDL_EventType:
        SDL_FIRSTEVENT
        SDL_QUIT
        SDL_APP_TERMINATING
        SDL_APP_LOWMEMORY
        SDL_APP_WILLENTERBACKGROUND
        SDL_APP_DIDENTERBACKGROUND
        SDL_APP_WILLENTERFOREGROUND
        SDL_APP_DIDENTERFOREGROUND
        SDL_DISPLAYEVENT
        SDL_WINDOWEVENT
        SDL_SYSWMEVENT
        SDL_KEYDOWN
        SDL_KEYUP
        SDL_TEXTEDITING
        SDL_TEXTINPUT
        SDL_KEYMAPCHANGED
        SDL_MOUSEMOTION
        SDL_MOUSEBUTTONDOWN
        SDL_MOUSEBUTTONUP
        SDL_MOUSEWHEEL
        SDL_JOYAXISMOTION
        SDL_JOYBALLMOTION
        SDL_JOYHATMOTION
        SDL_JOYBUTTONDOWN
        SDL_JOYBUTTONUP
        SDL_JOYDEVICEADDED
        SDL_JOYDEVICEREMOVED
        SDL_CONTROLLERAXISMOTION
        SDL_CONTROLLERBUTTONDOWN
        SDL_CONTROLLERBUTTONUP
        SDL_CONTROLLERDEVICEADDED
        SDL_CONTROLLERDEVICEREMOVED
        SDL_CONTROLLERDEVICEREMAPPED
        SDL_FINGERDOWN
        SDL_FINGERUP
        SDL_FINGERMOTION
        SDL_DOLLARGESTURE
        SDL_DOLLARRECORD
        SDL_MULTIGESTURE
        SDL_CLIPBOARDUPDATE
        SDL_DROPFILE
        SDL_DROPTEXT
        SDL_DROPBEGIN
        SDL_DROPCOMPLETE
        SDL_AUDIODEVICEADDED
        SDL_AUDIODEVICEREMOVED
        SDL_SENSORUPDATE
        SDL_RENDER_TARGETS_RESET
        SDL_RENDER_DEVICE_RESET
        SDL_USEREVENT
        SDL_LASTEVENT



# Copyright 2014 Tom Rothamel <tom@rothamel.us>
# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *

cdef Uint32 map_color(SDL_Surface *surface, color) except? 0xaabbccdd
cdef object get_color(Uint32 pixel, SDL_Surface *surface)
cdef to_sdl_color(color, SDL_Color *out)

cdef class Color:
    # Allow weak references.
    cdef object __weakref__

    cdef public Uint8 r, g, b, a
    cdef uint8_t length

    cdef from_rgba(self, Uint8 r, Uint8 g, Uint8 b, Uint8 a)
    cdef from_name(self, c)
    cdef from_hex(self, c)
